"""
Validate and clean request parameters for our endpoints using Django forms
"""
from django import forms
from django.conf import settings

from .defaults import FILTERS_HARDCODED
from .fields.array import ArrayField
from .fields.datetimerange import DatetimeRangeField


class CourseListForm(forms.Form):
    """
    Validate the query string params in the course list request
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Add in the fields for the custom filters from defaults/settings
        for filter_key in FILTERS_HARDCODED:
            self.fields[filter_key] = FILTERS_HARDCODED[filter_key]["field"](
                choices=[
                    (value, value) for value in FILTERS_HARDCODED[filter_key]["choices"]
                ],
                required=False,
            )

    end = DatetimeRangeField(required=False)
    enrollment_end = DatetimeRangeField(required=False)
    enrollment_start = DatetimeRangeField(required=False)
    languages = ArrayField(
        required=False, base_type=forms.ChoiceField(choices=settings.ALL_LANGUAGES)
    )
    limit = forms.IntegerField(required=False, min_value=1, initial=10)
    organizations = ArrayField(
        required=False, base_type=forms.IntegerField(min_value=0)
    )
    query = forms.CharField(required=False, min_length=3, max_length=200)
    offset = forms.IntegerField(required=False, min_value=0, initial=0)
    start = DatetimeRangeField(required=False)
    categories = ArrayField(required=False, base_type=forms.IntegerField(min_value=0))


class OrganizationListForm(forms.Form):
    """
    Validate the query string params in the organization list request
    """

    limit = forms.IntegerField(required=False, min_value=1, initial=10)
    query = forms.CharField(required=False, min_length=3, max_length=100)
    offset = forms.IntegerField(required=False, min_value=0, initial=0)


class CategoryListForm(forms.Form):
    """
    Validate the query string params in the category list request
    """

    limit = forms.IntegerField(required=False, min_value=1, initial=10)
    query = forms.CharField(required=False, min_length=3, max_length=100)
    offset = forms.IntegerField(required=False, min_value=0, initial=0)



"""Custom template tags for the courses application of Richie."""
from django import template

from classytags.arguments import Argument, MultiValueArgument
from cms.exceptions import PlaceholderNotFound
from cms.templatetags.cms_tags import (
    Placeholder,
    PlaceholderOptions,
    _get_page_by_untyped_arg,
)
from cms.toolbar.utils import get_toolbar_from_request
from cms.utils import get_site_id
from cms.utils.placeholder import validate_placeholder_name

# pylint: disable=invalid-name
register = template.Library()


class PagePlaceholder(Placeholder):
    """
    This template node is used to output page content and
    is also used in the admin to dynamically generate input fields.
    eg:
    {% page_placeholder "sidebar" page_lookup %}
    {% page_placeholder "sidebar" page_lookup inherit %}
    {% page_placeholder "sidebar" page_lookup as varname %}
    {% page_placeholder "sidebar" page_lookup or %}
        <div>No content</div>
    {% endpage_placeholder %}
    Keyword arguments:
        name: the name of the placeholder
        page_lookup: lookup argument for Page. See _get_page_by_untyped_arg() for detailed
            information on the allowed types and their interpretation for the page_lookup argument.
        inherit : optional argument which if given will result in inheriting
            the content of the placeholder with the same name on parent pages
        varname: context variable name. Output will be added to template context as this variable
            instead of being returned.
        or: optional argument which if given will make the template tag a block
            tag whose content is shown if the placeholder is empty
    """

    name = "page_placeholder"
    options = PlaceholderOptions(
        Argument("name", resolve=False),
        Argument("page_lookup"),
        MultiValueArgument("extra_bits", required=False, resolve=False),
        blocks=[("endpage_placeholder", "nodelist")],
    )

    # pylint: disable=arguments-differ,too-many-arguments
    def render_tag(self, context, name, page_lookup, extra_bits, nodelist=None):
        validate_placeholder_name(name)

        request = context.get("request")

        if request:
            page = _get_page_by_untyped_arg(page_lookup, request, get_site_id(None))

            toolbar = get_toolbar_from_request(request)
            renderer = toolbar.get_content_renderer()

            inherit = "inherit" in extra_bits

            # A placeholder is only editable on its own page
            editable = page == request.current_page

            try:
                content = renderer.render_page_placeholder(
                    slot=name,
                    context=context,
                    inherit=inherit,
                    page=page,
                    nodelist=nodelist,
                    editable=editable,
                )
            except PlaceholderNotFound:
                content = ""
        else:
            content = ""

        if not content and nodelist:
            content = nodelist.render(context)

        if "as" in extra_bits:
            try:
                varname = extra_bits[extra_bits.index("as") + 1]
            except IndexError:
                raise template.TemplateSyntaxError(
                    'the "as" word should be followed by the variable name'
                )
            context[varname] = content
            return ""

        return content


register.tag("page_placeholder", PagePlaceholder)



"""
ElasticSearch organization document management utilities
"""
from collections import defaultdict

from django.conf import settings
from django.db.models import Prefetch

from cms.models import Title
from djangocms_picture.models import Picture

from richie.plugins.simple_text_ckeditor.models import SimpleText

from ...courses.models import Organization
from .. import defaults
from ..exceptions import QueryFormatException
from ..forms import OrganizationListForm
from ..partial_mappings import MULTILINGUAL_TEXT
from ..utils.i18n import get_best_field_language
from ..utils.indexers import slice_string_for_completion


class OrganizationsIndexer:
    """
    Makes available the parameters the indexer requires as well as a function to shape
    objects into what we want to index in ElasticSearch
    """

    document_type = "organization"
    index_name = "richie_organizations"
    mapping = {
        "dynamic_templates": MULTILINGUAL_TEXT,
        "properties": {
            # Searchable
            **{
                "complete.{:s}".format(lang): {"type": "completion"}
                for lang, _ in settings.LANGUAGES
            },
            "description": {"type": "object"},
            "title": {"type": "object"},
            # Not searchable
            "absolute_url": {"type": "object", "enabled": False},
            "logo": {"type": "object", "enabled": False},
        },
    }
    scripts = {}
    display_fields = ["absolute_url", "logo", "title.*"]

    @classmethod
    def get_data_for_es(cls, index, action):
        """
        Load all the organizations from the Organization model and format them for the
        ElasticSearch index.
        """
        for organization in (
            Organization.objects.filter(
                extended_object__publisher_is_draft=False,
                extended_object__title_set__published=True,
            )
            .prefetch_related(
                Prefetch(
                    "extended_object__title_set",
                    to_attr="published_titles",
                    queryset=Title.objects.filter(published=True),
                )
            )
            .distinct()
        ):
            # Get published titles
            titles = {
                t.language: t.title
                for t in organization.extended_object.published_titles
            }

            # Get logo images
            logo_images = {}
            for logo_image in Picture.objects.filter(
                cmsplugin_ptr__placeholder__page=organization.extended_object,
                cmsplugin_ptr__placeholder__slot="logo",
            ):
                # Force the image format before computing it
                logo_image.use_no_cropping = False
                logo_image.width = defaults.ORGANIZATIONS_LOGO_IMAGE_WIDTH
                logo_image.height = defaults.ORGANIZATIONS_LOGO_IMAGE_HEIGHT
                logo_images[logo_image.cmsplugin_ptr.language] = logo_image.img_src

            # Get syllabus texts
            description = defaultdict(list)
            for simple_text in SimpleText.objects.filter(
                cmsplugin_ptr__placeholder__page=organization.extended_object,
                cmsplugin_ptr__placeholder__slot="description",
            ):
                description[simple_text.cmsplugin_ptr.language].append(simple_text.body)

            yield {
                "_id": str(organization.extended_object_id),
                "_index": index,
                "_op_type": action,
                "_type": cls.document_type,
                "absolute_url": {
                    language: organization.extended_object.get_absolute_url(language)
                    for language in titles.keys()
                },
                "complete": {
                    language: slice_string_for_completion(title)
                    for language, title in titles.items()
                },
                "logo": logo_images,
                "description": {l: " ".join(st) for l, st in description.items()},
                "title": titles,
            }

    @staticmethod
    def format_es_object_for_api(es_organization, best_language):
        """
        Format an organization stored in ES into a consistent and easy-to-consume record for
        API consumers
        """
        return {
            "id": es_organization["_id"],
            "logo": get_best_field_language(
                es_organization["_source"]["logo"], best_language
            ),
            "title": get_best_field_language(
                es_organization["_source"]["title"], best_language
            ),
        }

    @staticmethod
    def build_es_query(request):
        """
        Build an ElasticSearch query and its related aggregations, to be consumed by the ES client
        in the Organizations ViewSet
        """
        # Instantiate a form with the request's query_params to check & sanitize them
        query_params_form = OrganizationListForm(request.query_params)

        # Raise an exception with error information if the query params are not valid
        if not query_params_form.is_valid():
            raise QueryFormatException(query_params_form.errors)

        # Build a query that matches on the organization `title` field if it was passed by
        # the client
        # Note: test_elasticsearch_feature.py needs to be updated whenever the search call
        # is updated and makes use of new features.
        if query_params_form.cleaned_data.get("query"):
            query = {
                "query": {
                    "match": {
                        "title.fr": {
                            "query": query_params_form.cleaned_data.get("query"),
                            "analyzer": "french",
                        }
                    }
                }
            }
        # Build a match_all query by default
        else:
            query = {"query": {"match_all": {}}}

        return (
            query_params_form.cleaned_data.get("limit"),
            query_params_form.cleaned_data.get("offset") or 0,
            query,
        )



"""
Person CMS plugin
"""

from django.utils.translation import ugettext_lazy as _

from cms.plugin_base import CMSPluginBase
from cms.plugin_pool import plugin_pool

from .models import PersonPluginModel


@plugin_pool.register_plugin
class PersonPlugin(CMSPluginBase):
    """
    Person plugin displays a person's information on other pages
    """

    model = PersonPluginModel
    module = _("Persons")
    render_template = "persons/plugins/person.html"
    cache = True

    def render(self, context, instance, placeholder):
        context.update(
            {
                "instance": instance,
                "relevant_page": instance.relevant_page,
                "placeholder": placeholder,
            }
        )
        return context



"""
Settings specific to the courses application
"""
from django.conf import settings

PAGE_EXTENSION_TOOLBAR_ITEM_POSITION = getattr(
    settings, "RICHIE_PAGE_EXTENSION_TOOLBAR_ITEM_POSITION", 4
)



"""
Declare and configure the models for the courses application
"""
from collections import namedtuple

from django import forms
from django.conf import settings
from django.contrib.postgres.fields import ArrayField
from django.db import models
from django.utils import timezone, translation
from django.utils.functional import lazy
from django.utils.translation import ugettext_lazy as _

from cms.api import Page
from cms.extensions.extension_pool import extension_pool
from cms.models.pluginmodel import CMSPlugin
from filer.fields.image import FilerImageField

from ...core.fields.multiselect import MultiSelectField
from ...core.models import BasePageExtension, PagePluginMixin
from .category import Category
from .organization import Organization

CourseState = namedtuple("CourseState", ["priority", "cta", "text", "datetime"])


class ChoiceArrayField(ArrayField):
    """
    A field that allows us to store an array of choices.
    Uses Django's Postgres ArrayField
    and a MultipleChoiceField for its formfield.
    """

    def formfield(self, **kwargs):
        defaults = {
            "form_class": forms.MultipleChoiceField,
            "choices": self.base_field.choices,
        }
        defaults.update(kwargs)
        # Skip our parent's formfield implementation completely as we don't
        # care for it.
        # pylint:disable=bad-super-call
        return super(ArrayField, self).formfield(**defaults)


class Course(BasePageExtension):
    """
    The course page extension represents and records a course in the catalog.

    This model should be used to record structured data about the course whereas the
    associated page object is where we record the less structured information to display on the
    page that presents the course.
    """

    ROOT_REVERSE_ID = "courses"
    TEMPLATE_DETAIL = "courses/cms/course_detail.html"

    class Meta:
        verbose_name = _("course")

    def __str__(self):
        """Human representation of a course."""
        return "{model}: {title}".format(
            model=self._meta.verbose_name.title(),
            title=self.extended_object.get_title(),
        )

    def get_organizations(self):
        """
        Return the organizations linked to the course via an organization plugin in the
        placeholder `course_organizations` on the course detail page, ranked by their
        `position`.
        """
        selector = "extended_object__organization_plugins__cmsplugin_ptr__placeholder"
        # pylint: disable=no-member
        filter_dict = {
            "{:s}__page".format(selector): self.extended_object,
            "{:s}__slot".format(selector): "course_organizations",
        }
        # For a public course, we must filter out organizations that are not published in
        # any language
        if self.extended_object.publisher_is_draft is False:
            filter_dict["extended_object__title_set__published"] = True

        return (
            Organization.objects.filter(**filter_dict)
            .select_related("extended_object")
            .order_by("extended_object__organization_plugins__cmsplugin_ptr__position")
            .distinct()
        )

    def get_main_organization(self):
        """
        Return the main organization linked to the course via an organization plugin in the
        placeholder `course_organizations` on the course detail page.

        Plugins are sortable by drag-and-drop in the plugin bar so we use this "position"
        information to return the first one as the main organization.
        """
        return self.get_organizations().first()

    def get_categories(self):
        """
        Return the categories linked to the course via a category plugin in the placeholder
        `course_categories` on the course detail page, ranked by their `position`.
        """
        selector = "extended_object__category_plugins__cmsplugin_ptr__placeholder"
        # pylint: disable=no-member
        filter_dict = {
            "{:s}__page".format(selector): self.extended_object,
            "{:s}__slot".format(selector): "course_categories",
        }
        # For a public course, we must filter out categories that are not published in
        # any language
        if self.extended_object.publisher_is_draft is False:
            filter_dict["extended_object__title_set__published"] = True

        return (
            Category.objects.filter(**filter_dict)
            .select_related("extended_object")
            .order_by("extended_object__category_plugins__cmsplugin_ptr__position")
            .distinct()
        )

    def get_root_to_leaf_category_pages(self):
        """
        Build a query retrieving all pages linked to the category or one of its ancestors,
        excluding the meta category itself.

        This is useful to build the course Elasticsearch index. When category is linked to a
        course, we associate all its ancestors (excluding the meta category) to the course so
        that a search for courses in an ancestor category also returns the course.
        """
        category_query = self.get_categories()

        # 1. We want the pages directly related to a category of the course
        page_query = Page.objects.filter(publisher_draft__category__in=category_query)
        # 2. We want the pages related to one of the ancestors of the categories of the course
        for category in category_query.select_related(
            "public_extension__extended_object"
        ):
            ancestor_query = (
                category.public_extension.extended_object.get_ancestor_pages()
            )
            # Don't include the meta category as it materializes a "filter bank" and not a
            # search option
            page_query = page_query | ancestor_query.filter(
                node__parent__cms_pages__category__isnull=False
            )
        return page_query.distinct()

    def get_course_runs(self):
        """
        Returns a query yielding the course runs related to the course. They may be direct
        children of the course or children of a snapshot of the course.

        For a draft course instance, the related draft course runs are retrieved.
        For a public course instance, the related public course runs are retrieved, but only
        those that are currently published in at least one language*.

        (*) The catch here is that a course run could have been published and then unpublished
            for all languages. The public instance is created the first time a draft page is
            published. Of course, this public instance will still exist if the object is then
            unpublished for all languages...
            Said differently, a page can have a public version of itself in database but not be
            currently published in any language.
        """
        node = self.extended_object.node
        is_draft = self.extended_object.publisher_is_draft
        filter_dict = {
            "extended_object__publisher_is_draft": is_draft,
            "extended_object__node__path__startswith": node.path,
            "extended_object__node__depth__gt": node.depth,
        }
        # For a public course, we must filter out course runs that are not published in
        # any language
        if is_draft is False:
            filter_dict["extended_object__title_set__published"] = True

        return (
            CourseRun.objects.filter(**filter_dict)
            .order_by("extended_object__node__path")
            .distinct()
        )

    def get_course_runs_for_language(self, language=None):
        """
        Returns a query yielding the course runs related to the course for the current
        language (or the language passed in arguments). They may be direct children of
        the course or children of a snapshot of the course.
        """
        language = language or translation.get_language()
        course_runs = self.get_course_runs()

        # For a public course, we must filter out course runs that are not published in
        # the language
        if self.extended_object.publisher_is_draft is False:
            course_runs = course_runs.filter(
                extended_object__title_set__language=language,
                extended_object__title_set__published=True,
            )
        else:
            course_runs = course_runs.filter(
                extended_object__title_set__language=language
            )

        return course_runs

    @property
    def state(self):
        """
        The state of the course carrying information on what to display on a course glimpse.

        The game is to find, in the correct order, the first of the following conditions that is
        verified for this course:
          0: a run is on-going and open for enrollment > "closing on": {enrollment_end}
          1: a run is future and open for enrollment > "starting on": {start}
          2: a run is future and not yet open for enrollment > "starting on": {start}
          3: a run is future and no more open for enrollment > "closed": {None}
          4: a run is on-going but closed for enrollment > "on going": {None}
          5: there's a finished run in the past > "archived": {None}
          6: there are no runs at all > "coming soon": {None}
        """
        # The default state is for a course that has no course runs
        best_state = CourseState(6, None, _("coming soon"), None)

        for course_run in self.get_course_runs_for_language().only(
            "start", "end", "enrollment_start", "enrollment_end"
        ):
            state = course_run.state
            if state.priority < best_state.priority:
                best_state = state
            if state.priority == 0:
                # We found the best state, don't waste more time
                break

        return best_state

    def save(self, *args, **kwargs):
        """
        Enforce validation each time an instance is saved
        """
        self.full_clean()
        super().save(*args, **kwargs)


class CourseRun(BasePageExtension):
    """
    The course run represents and records the occurence of a course between a start
    and an end date.
    """

    resource_link = models.URLField(_("Resource link"), blank=True, null=True)
    start = models.DateTimeField(_("course start"), blank=True, null=True)
    end = models.DateTimeField(_("course end"), blank=True, null=True)
    enrollment_start = models.DateTimeField(
        _("enrollment start"), blank=True, null=True
    )
    enrollment_end = models.DateTimeField(_("enrollment end"), blank=True, null=True)
    languages = MultiSelectField(
        max_choices=50,
        max_length=255,  # MySQL does not allow max_length > 255
        # Language choices are made lazy so that we can override them in our tests.
        # When set directly, they are evaluated too early and can't be changed with the
        # "override_settings" utility.
        choices=lazy(lambda: settings.ALL_LANGUAGES, tuple)(),
        help_text=_("The list of languages in which the course content is available."),
    )

    TEMPLATE_DETAIL = "courses/cms/course_run_detail.html"

    class Meta:
        verbose_name = _("course run")

    def __str__(self):
        """Human representation of a course run."""
        start = "{:%y/%m/%d %H:%M} - ".format(self.start) if self.start else ""
        return "{start:s}{course:s}".format(
            course=self.extended_object.get_title(), start=start
        )

    # pylint: disable=arguments-differ
    def save(self, *args, **kwargs):
        """
        Enforce validation each time an instance is saved.
        """
        self.full_clean()
        super().save(*args, **kwargs)

    @staticmethod
    def compute_state(start, end, enrollment_start, enrollment_end):
        """
        Compute at the current time the state of a course run that would have the dates
        passed in argument.

        A static method not using the instance allows to call it with an Elasticsearch result.

        Several states are possible for a course run each of which is given a priority. The
        lower the priority, the more interesting the course run is (a course run open for
        enrollment is more interesting than an archived course run):
          0: a run is on-going and open for enrollment > "closing on": {enrollment_end}
          1: a run is future and open for enrollment > "starting on": {start}
          2: a run is future and not yet open for enrollment > "starting on": {start}
          3: a run is future and no more open for enrollment > "closed": {None}
          4: a run is on-going but closed for enrollment > "on going": {None}
          5: there's a finished run in the past > "archived": {None}
          6: there are no runs at all > "coming soon": {None}
        """
        now = timezone.now()
        if start < now:
            if end > now:
                if enrollment_end > now:
                    # ongoing open
                    return CourseState(
                        0, _("enroll now"), _("closing on"), enrollment_end
                    )
                # ongoing closed
                return CourseState(4, None, _("on-going"), None)
            # archived
            return CourseState(5, None, _("archived"), None)
        if enrollment_start > now:
            # future not yet open
            return CourseState(2, _("see details"), _("starting on"), start)
        if enrollment_end > now:
            # future open
            return CourseState(1, _("enroll now"), _("starting on"), start)
        # future already closed
        return CourseState(3, None, _("enrollment closed"), None)

    @property
    def state(self):
        """Return the state of the course run at the current time."""
        return self.compute_state(
            self.start, self.end, self.enrollment_start, self.enrollment_end
        )


class CoursePluginModel(PagePluginMixin, CMSPlugin):
    """
    Course plugin model handles the relation from CoursePlugin
    to their Course instance
    """

    page = models.ForeignKey(
        Page,
        on_delete=models.CASCADE,
        related_name="course_plugins",
        limit_choices_to={
            # There's a draft and a public course attached to the draft and
            # the public parent. Without the first condition, 4 options are
            # availables for each course. The second condition makes sure the
            # parent is not a course. The third option filters out public
            # course. The fourth option makes sure only courses show up.
            # If there's is no parent, the course is filtered out.
            "node__parent__cms_pages__publisher_is_draft": True,
            "node__parent__cms_pages__course__isnull": True,
            "publisher_is_draft": True,
            "course__isnull": False,
        },
    )

    class Meta:
        verbose_name = _("course plugin model")

    def __str__(self):
        """Human representation of a page plugin"""
        return "{model:s}: {id:d}".format(
            model=self._meta.verbose_name.title(), id=self.id
        )


class Licence(models.Model):
    """
    Licence model.

    Instances of this models should only be created by administrators.
    """

    name = models.CharField(_("name"), max_length=200)
    logo = FilerImageField(
        verbose_name=_("logo"), on_delete=models.PROTECT, related_name="licence"
    )
    url = models.CharField(_("url"), blank=True, max_length=255)
    content = models.TextField(_("content"), blank=False, default="")

    class Meta:
        verbose_name = _("licence")

    def __str__(self):
        """Human representation of a person title"""
        return "{model}: {name}".format(
            model=self._meta.verbose_name.title(), name=self.name
        )


class LicencePluginModel(CMSPlugin):
    """
    Licence plugin model.
    """

    licence = models.ForeignKey(Licence, on_delete=models.CASCADE)
    description = models.TextField(_("description"), blank=True, default="")

    def __str__(self):
        """Human representation of a person plugin"""
        return "{model:s}: {name:s}".format(
            model=self._meta.verbose_name.title(), name=self.licence.name
        )


extension_pool.register(Course)
extension_pool.register(CourseRun)



"""
Courses application admin
"""
import time

from django.conf.urls import url
from django.contrib import admin
from django.db import models, transaction
from django.http import HttpResponseBadRequest, HttpResponseForbidden, JsonResponse
from django.utils.decorators import method_decorator
from django.utils.encoding import force_text
from django.utils.translation import ugettext_lazy as _
from django.views.decorators.http import require_POST

from cms.admin.placeholderadmin import FrontendEditableAdminMixin
from cms.api import Page, create_title
from cms.extensions import PageExtensionAdmin
from cms.utils import page_permissions
from cms.utils.admin import jsonify_request

from .fields import CourseRunSplitDateTimeField
from .forms import LicenceFormAdmin
from .models import Course, CourseRun, Licence, Organization
from .widgets import CourseRunSplitDateTimeWidget

REQUIRE_POST = method_decorator(require_POST)


class CourseAdmin(PageExtensionAdmin):
    """Admin class for the Course model"""

    list_display = ["title"]

    # pylint: disable=no-self-use
    def title(self, obj):
        """
        Display the course title as a read-only field from the related page
        """
        return obj.extended_object.get_title()

    def get_urls(self):
        """
        Add admin URL to trigger the snapshot of a course.
        """
        url_patterns = super().get_urls()
        return [
            url(
                r"^(?P<course_id>[0-9]+)/snapshot/$",
                self.admin_site.admin_view(self.snapshot),
                name="cms_course_snapshot",
            )
        ] + url_patterns

    @REQUIRE_POST
    @transaction.atomic
    def snapshot(self, request, course_id, *args, **kwargs):
        """
        Snapshotting a course is making a copy of the course page with all its permissions and
        extensions, and placing the copy as the first child of the page being snapshotted, then
        moving all the course run of the page being snapshotted as children of the new snapshot
        so we keep record of the course as it was when these course runs were played.
        """
        try:
            page = Page.objects.select_related("node__site").get(
                publisher_is_draft=True, course__id=course_id
            )
        except Page.DoesNotExist:
            return jsonify_request(
                HttpResponseBadRequest(
                    force_text(_("Error! Course could not be found."))
                )
            )

        # If the page has a parent that is a course page, it is a snapshot and should therefore
        # not be allowed to be itself snapshotted.
        if page.parent_page:
            try:
                page.parent_page.course
            except Course.DoesNotExist:
                pass
            else:
                return jsonify_request(
                    HttpResponseForbidden(
                        force_text(_("Error! You can't snapshot a snapshot."))
                    )
                )

        site = page.node.site

        # User can only snapshot pages he can see
        can_snapshot = page_permissions.user_can_change_page(request.user, page, site)

        if can_snapshot:
            # User can only snapshot a page if he has the permission to add a page under it.
            can_snapshot = page_permissions.user_can_add_subpage(
                request.user, page, site
            )

        if not can_snapshot:
            return jsonify_request(
                HttpResponseForbidden(
                    force_text(
                        _("Error! You don't have permissions to snapshot this page.")
                    )
                )
            )

        # Copy the page as its own child with its extension.
        # Titles are set to a timestamp in each language of the original page
        new_page = page.copy(
            site, parent_node=page.node, translations=False, extensions=True
        )

        # The snapshot title and slug is set to a timestamp of the time of snapshot. It is
        # published only in languages for which the original course page was published.
        for language in page.get_languages():
            base = page.get_path(language)
            timestamp = str(int(time.time()))
            snapshot_title = _("Snapshot of {:s}").format(page.get_title(language))
            create_title(
                language=language,
                menu_title=timestamp,
                title="{:s} - {:s}".format(timestamp, snapshot_title),
                slug=timestamp,
                path="{:s}/{:s}".format(base, timestamp) if base else timestamp,
                page=new_page,
            )
            if page.is_published(language) is True:
                new_page.publish(language)

        # Move the existing course run subpages as children of the snapshot
        # Their publication status will be respected
        for subpage in page.get_child_pages().filter(courserun__isnull=False):
            subpage.move_page(new_page.node, position="last-child")

        return JsonResponse({"id": new_page.course.id})


class CourseRunAdmin(FrontendEditableAdminMixin, PageExtensionAdmin):
    """Admin class for the CourseRun model"""

    list_display = ["title"]
    frontend_editable_fields = (
        "languages",
        "resource_link",
        "start",
        "end",
        "enrollment_start",
        "enrollment_end",
    )
    formfield_overrides = {
        models.DateTimeField: {
            "form_class": CourseRunSplitDateTimeField,
            "widget": CourseRunSplitDateTimeWidget,
        }
    }

    # pylint: disable=no-self-use
    def title(self, obj):
        """
        Display the course title as a read-only field from the related page
        """
        return obj.extended_object.get_title()


class OrganizationAdmin(PageExtensionAdmin):
    """
    Admin class for the Organization model
    """

    list_display = ["title", "code"]

    # pylint: disable=no-self-use
    def title(self, obj):
        """
        Get the page title from the related page
        """
        return obj.extended_object.get_title()


class LicenceAdmin(admin.ModelAdmin):
    """
    Admin class for the Licence model
    """

    list_display = ["name"]
    form = LicenceFormAdmin


admin.site.register(Course, CourseAdmin)
admin.site.register(CourseRun, CourseRunAdmin)
admin.site.register(Organization, OrganizationAdmin)
admin.site.register(Licence, LicenceAdmin)



"""
Declare and configure the models for the courses application
"""
from django.apps import apps
from django.conf import settings
from django.db import models
from django.db.models import Prefetch
from django.utils import translation
from django.utils.translation import ugettext_lazy as _

from cms.api import Page
from cms.extensions.extension_pool import extension_pool
from cms.models import Title
from cms.models.pluginmodel import CMSPlugin

from ...core.models import BasePageExtension, PagePluginMixin


class Category(BasePageExtension):
    """
    The category page extension represents and records a thematic in the catalog.

    This model should be used to record structured data about the thematic whereas the
    associated page object is where we record the less structured information to display on the
    page that presents the thematic.
    """

    ROOT_REVERSE_ID = "categories"
    TEMPLATE_DETAIL = "courses/cms/category_detail.html"

    class Meta:
        verbose_name = _("category")
        ordering = ["-pk"]

    def __str__(self):
        """Human representation of a category"""
        ancestors = self.extended_object.get_ancestor_pages().filter(
            category__isnull=False
        )
        return "{ancestors:s}{title:s}".format(
            ancestors="{:s} / ".format(" / ".join([a.get_title() for a in ancestors]))
            if ancestors
            else "",
            title=self.extended_object.get_title(),
        )

    def get_courses(self, language=None):
        """
        Return a query to get the courses related to this category ie for which a plugin for
        this category is linked to the course page on the "course_categories" placeholder.
        """
        page = (
            self.extended_object
            if self.extended_object.publisher_is_draft
            else self.draft_extension.extended_object
        )
        language = language or translation.get_language()
        bfs = "extended_object__placeholders__cmsplugin__courses_categorypluginmodel__page"
        filter_dict = {
            "extended_object__node__parent__cms_pages__course__isnull": True,
            "extended_object__publisher_is_draft": True,
            "extended_object__placeholders__slot": "course_categories",
            "extended_object__placeholders__cmsplugin__language": language,
            bfs: page,
            "{:s}__publisher_is_draft".format(bfs): True,
        }
        course_model = apps.get_model(app_label="courses", model_name="course")
        # pylint: disable=no-member
        return (
            course_model.objects.filter(**filter_dict)
            .select_related("extended_object")
            .prefetch_related(
                Prefetch(
                    "extended_object__title_set",
                    to_attr="prefetched_titles",
                    queryset=Title.objects.filter(language=language),
                )
            )
            .distinct()
        )


def get_category_limit_choices_to():
    """Return a query limiting the categories proposed when creating a CategoryPlugin."""
    limit_choices_to = {
        "publisher_is_draft": True,  # plugins work with draft instances
        "category__isnull": False,  # limit to pages linked to a category object
        "node__parent__cms_pages__category__isnull": False,  # exclude meta categories
    }
    # Limit to leaf categories only if active in settings (False by default)
    if getattr(settings, "LIMIT_PLUGIN_CATEGORIES_TO_LEAF", False):
        limit_choices_to["node__numchild"] = 0

    return limit_choices_to


class CategoryPluginModel(PagePluginMixin, CMSPlugin):
    """
    Category plugin model handles the relation between CategoryPlugin
    and their Category instance

    We only propose to link category objects:
    - are not at the root of the category tree (this position defines a meta category which
      is used to define a filter bank on the search page e.g. Subjects, Levels).
    - that are a leaf in the category tree (can be deactivated with setting
      `LIMIT_PLUGIN_CATEGORIES_TO_LEAF`),
    """

    page = models.ForeignKey(
        Page,
        related_name="category_plugins",
        limit_choices_to=get_category_limit_choices_to,
        on_delete=models.CASCADE,
    )

    class Meta:
        verbose_name = _("category plugin model")

    def __str__(self):
        """Human representation of a page plugin"""
        return "{model:s}: {id:d}".format(
            model=self._meta.verbose_name.title(), id=self.id
        )


extension_pool.register(Category)



"""
Large banner plugin field choices
"""
from django.conf import settings
from django.utils.translation import ugettext_lazy as _

LARGEBANNER_TEMPLATES = getattr(
    settings,
    "RICHIE_LARGEBANNER_TEMPLATES",
    [("richie/large_banner/large_banner.html", _("Default"))],
)



"""
CMS Wizard to add a person page
"""
from django import forms
from django.template.defaultfilters import slugify
from django.utils.functional import cached_property
from django.utils.translation import get_language
from django.utils.translation import ugettext_lazy as _

from cms.api import create_page
from cms.forms.wizards import SlugWidget
from cms.models import Page
from cms.wizards.wizard_base import Wizard
from cms.wizards.wizard_pool import wizard_pool

from .models import Person, PersonTitle


class BaseWizardForm(forms.Form):
    """
    Factorize the part of the wizard form that handles the title and the slug as they are
    common to all our wizard forms.
    """

    title = forms.CharField(
        max_length=255,
        required=True,
        widget=forms.TextInput(),
        label=_("Page title"),
        help_text=_("Title of the page in current language"),
    )
    slug = forms.CharField(
        max_length=200,  # Should be less than 255, the "max_length" of a Page's "path" field
        required=False,
        widget=SlugWidget(),
        label=_("Page slug"),
        help_text=_("Slug of the page in current language"),
    )

    def clean(self):
        """
        Ensure that the slug field is set or derive it from the title
        """
        cleaned_data = super().clean()

        # If the slug is not explicitly set, generate it from the title
        if cleaned_data.get("title") and not cleaned_data.get("slug"):
            cleaned_data["slug"] = slugify(cleaned_data["title"])[:200]

        # Check that the length of the slug is compatible with its parent page:
        #  a page `path` is limited to 255 chars, therefore the course page slug should
        # always be shorter than (255 - length of parent page path - 1 character for the "/")
        if self.parent_page:
            length = len(self.parent_page.get_path()) + 1 + len(cleaned_data["slug"])
            if length > 255:
                raise forms.ValidationError(
                    {
                        "slug": [
                            _(
                                "This slug is too long. The length of the path built by "
                                "prepending the slug of the parent page would be {:d} characters "
                                "long and it should be less than 255".format(length)
                            )
                        ]
                    }
                )

        return cleaned_data

    @cached_property
    def parent_page(self):
        """
        The parent page for each type of page is defined on its page extension model.
        """
        try:
            return Page.objects.get(
                reverse_id=self.model.ROOT_REVERSE_ID, publisher_is_draft=True
            )
        except Page.DoesNotExist:
            raise forms.ValidationError(
                {
                    "slug": [
                        _(
                            "You must first create a parent page and set its `reverse_id` to "
                            "`{reverse}`.".format(reverse=self.model.ROOT_REVERSE_ID)
                        )
                    ]
                }
            )

    def save(self):
        """
        Create the page with "title" and "slug"
        """
        return create_page(
            title=self.cleaned_data["title"],
            slug=self.cleaned_data["slug"],
            language=get_language(),
            parent=self.parent_page,
            template=self.model.TEMPLATE_DETAIL,
            published=False,  # The creation wizard should not publish the page
        )


class PersonWizardForm(BaseWizardForm):
    """
    This form is used by the wizard that creates a new person page.
    A related Person model is created for each person page.
    """

    person_title = forms.ModelChoiceField(
        required=True,
        queryset=PersonTitle.objects.all(),
        label=_("Title"),
        help_text=_("Choose this person's title among existing ones"),
    )
    first_name = forms.CharField(
        max_length=200, widget=forms.TextInput(), label=_("First Name")
    )
    last_name = forms.CharField(
        max_length=200, widget=forms.TextInput(), label=_("Last Name")
    )

    model = Person

    def save(self):
        """
        The parent form created the page.
        This method creates the associated person page extension.
        """
        page = super().save()
        Person.objects.create(
            extended_object=page,
            person_title=self.cleaned_data["person_title"],
            first_name=self.cleaned_data["first_name"],
            last_name=self.cleaned_data["last_name"],
        )
        return page


class PersonWizard(Wizard):
    """Inherit from Wizard because each wizard must have its own Python class."""


wizard_pool.register(
    PersonWizard(
        title=_("New person page"),
        description=_("Create a new person page"),
        model=Person,
        form=PersonWizardForm,
        weight=200,
    )
)



"""
Utils common to all applications of the Richie CMS.
"""
import io
import os
import random


def file_getter(image_type):
    """
    This function can be passed to factory boy ImageField to dynamically generate
    image fields.

    It randomly picks-up an image from targeted directory (core fixtures directory +
    image type dir).

    Arguments:
        image_type (string): Directory name in fixtures directory that contains some images to
            pick up.

    Returns:
        callable: A function that picks a random image for a given type of image (logo,
        banner,...)
    """

    def pick_random(filename=None):
        """
        Pick a random file from fixtures within the image type passed as argument to the parent
        function.
        """
        image_directory = os.path.join(
            os.path.dirname(__file__), "fixtures", image_type
        )
        filename = filename or random.choice(os.listdir(image_directory))

        # Factory boy's "from_func" param is expecting a file but does not seem to close it
        # properly. Let's load the content of the file in memory and pass it as a BytesIO to
        # factory boy so that the file is nicely closed
        with open(os.path.join(image_directory, filename), "rb") as image_file:
            in_memory_file = io.BytesIO(image_file.read())
            in_memory_file.name = filename
        return in_memory_file

    return pick_random



"""
Common utilities related to our indexers. For use in our indexers and related settings,
or as helpers for users of the project.
"""
from collections import namedtuple

# Define a named tuple type that will enforce the necessary keys for our ES_INDICES setting
# (and also make iteration, both with and without keys, trivial)
IndicesList = namedtuple("IndicesList", ["courses", "organizations", "categories"])


def slice_string_for_completion(string):
    """
    Split a string in significant parts for use in completion.
    Example: "University of Paris 13" => "University of Paris 13", "of Paris 13", "Paris 13", "13"

    This is useful to enable autocompletion starting from any part of a name. If we just use the
    name directly in the ES completion type, it will only return options that match on the first
    characters of the whole string, which is not always suitable.
    """
    parts = [part for part in string.split(" ") if part != ""]
    return [" ".join(parts[index:]) for index, _ in enumerate(parts)]



"""
CMS Wizard to add a course page
"""
from django import forms
from django.conf import settings
from django.utils.functional import cached_property
from django.utils.translation import get_language
from django.utils.translation import ugettext_lazy as _

from cms.api import add_plugin
from cms.wizards.wizard_base import Wizard
from cms.wizards.wizard_pool import wizard_pool

from richie.apps.persons.cms_wizards import BaseWizardForm

from .models import Category, Course, CourseRun, Organization


class CourseWizardForm(BaseWizardForm):
    """
    This form is used by the wizard that creates a new course page.
    A related Course model is created for each course page.
    """

    organization = forms.ModelChoiceField(
        required=True,
        queryset=Organization.objects.filter(extended_object__publisher_is_draft=True),
        label=_("Organization"),
        help_text=_("The organization in charge of this course"),
    )

    model = Course

    def save(self):
        """
        The parent form created the page.
        This method creates the associated course page extension.
        """
        page = super().save()
        course = Course.objects.create(extended_object=page)
        # Add a plugin for the organization

        placeholder = course.extended_object.placeholders.get(
            slot="course_organizations"
        )
        add_plugin(
            language=get_language(),
            placeholder=placeholder,
            plugin_type="OrganizationPlugin",
            **{"page": self.cleaned_data["organization"].extended_object},
        )

        return page


class CourseWizard(Wizard):
    """Inherit from Wizard because each wizard must have its own Python class."""


wizard_pool.register(
    CourseWizard(
        title=_("New course page"),
        description=_("Create a new course page"),
        model=Course,
        form=CourseWizardForm,
        weight=200,
    )
)


class CourseRunWizardForm(BaseWizardForm):
    """
    This form is used by the wizard that creates a new course run page.
    A related CourseRun model is created for each course run page.
    """

    course = forms.ModelChoiceField(
        required=True,
        queryset=Course.objects.filter(
            extended_object__publisher_is_draft=True,
            # Only list the top level courses (not the snapshots)
            extended_object__node__parent__cms_pages__course__isnull=True,
        ).distinct(),
        label=_("Course"),
        help_text=_("The course that this course run describes."),
    )
    languages = forms.MultipleChoiceField(
        required=True,
        label=_("Languages"),
        choices=settings.ALL_LANGUAGES,
        help_text=_(
            "Select all the languages in which the course content is available."
        ),
    )
    resource_link = forms.URLField(label=_("Resource link"), required=False)

    model = CourseRun

    @cached_property
    def parent_page(self):
        """
        The parent page of a course run is the related course.
        """
        course = self.cleaned_data.get("course")
        return course.extended_object if course else None

    def save(self):
        """
        The parent form created the page.
        This method creates the associated course run page extension.
        """
        page = super().save()
        CourseRun.objects.create(
            extended_object=page,
            resource_link=self.cleaned_data["resource_link"],
            languages=self.cleaned_data["languages"],
        )
        return page


class CourseRunWizard(Wizard):
    """Inherit from Wizard because each wizard must have its own Python class."""


wizard_pool.register(
    CourseRunWizard(
        title=_("New course run page"),
        description=_("Create a new course run page"),
        model=CourseRun,
        form=CourseRunWizardForm,
        weight=200,
    )
)


class OrganizationWizardForm(BaseWizardForm):
    """
    This form is used by the wizard that creates a new organization page
    A related organization model is created for each organization page
    """

    model = Organization

    def save(self):
        """
        The parent form created the page.
        This method creates the associated organization.
        """
        page = super().save()
        Organization.objects.create(extended_object=page)
        return page


class OrganizationWizard(Wizard):
    """Inherit from Wizard because each wizard must have its own Python class."""


wizard_pool.register(
    OrganizationWizard(
        title=_("New Organization page"),
        description=_("Create a new Organization page"),
        model=Organization,
        form=OrganizationWizardForm,
        weight=200,
    )
)


class CategoryWizardForm(BaseWizardForm):
    """
    This form is used by the wizard that creates a new category page.
    A related Category model is created for each category page.
    """

    parent_category = forms.ModelChoiceField(
        required=False,
        queryset=Category.objects.filter(
            extended_object__publisher_is_draft=True
        ).order_by("extended_object__node__path"),
        label=_("Parent category"),
        help_text=_("Choose a parent if you are building a category tree."),
    )

    model = Category

    @cached_property
    def parent_page(self):
        """
        The parent page may be defined in the form but defaults to the categories root page
        defined by the `ROOT_REVERSE_ID` property of the Category model.
        """
        parent_category = self.cleaned_data.get("parent_category")
        return (
            parent_category.extended_object if parent_category else super().parent_page
        )

    def save(self):
        """
        The parent form created the page.
        This method creates the associated category page extension.
        """
        page = super().save()
        Category.objects.create(extended_object=page)
        return page


class CategoryWizard(Wizard):
    """Inherit from Wizard because each wizard must have its own Python class."""


wizard_pool.register(
    CategoryWizard(
        title=_("New category page"),
        description=_("Create a new category page"),
        model=Category,
        form=CategoryWizardForm,
        weight=200,
    )
)



"""
Centralize our custom ES mappings for easier reuse and testability
"""

# Dynamic template to match translated strings with their language analyzer
# NB: expects objects to follow a conventional form, eg:
#   {
#       'title': {
#           'en': '<english title>',
#           'fr': '<french title>',
#       },
#   }
# Where all the language keys are optional and dynamically added to the mapping at indexing time
MULTILINGUAL_TEXT = [
    {
        "english_strings": {
            "match_mapping_type": "string",
            "path_match": "*.en",
            "mapping": {"type": "text", "analyzer": "english"},
        }
    },
    {
        "french_strings": {
            "match_mapping_type": "string",
            "path_match": "*.fr",
            "mapping": {"type": "text", "analyzer": "french"},
        }
    },
]



"""
Persons factories
"""
from django.core.files import File
from django.utils.translation import get_language

import factory
from cms.api import add_plugin
from filer.models.imagemodels import Image

from ..core.factories import PageExtensionDjangoModelFactory
from ..core.helpers import create_text_plugin
from ..core.utils import file_getter
from .models import Person, PersonTitle, PersonTitleTranslation


class PersonTitleFactory(factory.django.DjangoModelFactory):
    """
    Factory to generate random yet realistic PersonTitle objects with one default translation.
    """

    translation = factory.RelatedFactory(
        "richie.apps.persons.factories.PersonTitleTranslationFactory", "master"
    )

    class Meta:
        model = PersonTitle


class PersonTitleTranslationFactory(factory.django.DjangoModelFactory):
    """
    Factory to generate random yet realistic translation instances for the PersonTitle model.
    """

    class Meta:
        model = PersonTitleTranslation

    master = None
    language_code = factory.LazyAttribute(lambda o: get_language())
    title = factory.Faker("prefix")
    abbreviation = factory.LazyAttribute(lambda o: o.title[0])


class PersonFactory(PageExtensionDjangoModelFactory):
    """
    Person factory to generate random yet realistic person's name and title
    """

    class Meta:
        model = Person
        exclude = [
            "page_in_navigation",
            "page_languages",
            "page_parent",
            "page_template",
            "page_title",
        ]

    # fields concerning the related page
    page_template = Person.TEMPLATE_DETAIL

    first_name = factory.Faker("first_name")
    last_name = factory.Faker("last_name")
    person_title = factory.SubFactory(PersonTitleFactory)

    @factory.lazy_attribute
    def page_title(self):
        """
        Build the page title from the person's title and names
        """
        names = [
            self.person_title.title if self.person_title else None,
            self.first_name,
            self.last_name,
        ]
        # Join the names that are null into a string
        return " ".join([n for n in names if n is not None])

    @factory.post_generation
    # pylint: disable=unused-argument
    def fill_portrait(self, create, extracted, **kwargs):
        """
        Add a portrait with a random image
        """
        if create and extracted:
            portrait_placeholder = self.extended_object.placeholders.get(
                slot="portrait"
            )

            portrait_file = file_getter("portrait")()
            wrapped_portrait = File(portrait_file, portrait_file.name)
            portrait = Image.objects.create(file=wrapped_portrait)
            for language in self.extended_object.get_languages():
                add_plugin(
                    language=language,
                    placeholder=portrait_placeholder,
                    plugin_type="PicturePlugin",
                    picture=portrait,
                    attributes={"alt": "portrait image"},
                )

    @factory.post_generation
    # pylint: disable=unused-argument
    def fill_resume(self, create, extracted, **kwargs):
        """
        Add a text plugin for resume with a long random text
        """
        if create and extracted:
            create_text_plugin(
                self.extended_object,
                "resume",
                nb_paragraphs=1,
                languages=self.extended_object.get_languages(),
            )



"""
Helpers that can be useful throughout the whole project
"""
import random

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.utils.text import slugify

import factory
from cms.api import add_plugin, create_page, create_title


def create_i18n_page(title=None, languages=None, is_homepage=False, **kwargs):
    """
    Creating a multilingual page is not straightforward so we should have a helper

    This content argument should be a dictionary with the title of the page in each language:

        {
            'en': 'About',
            'fr': 'A propos',
            'de': 'Impressum',
        }

    """
    template = kwargs.pop("template", None) or "richie/fullwidth.html"

    if title is None:
        # Create realistic titles in each language with faker
        languages = languages or [settings.LANGUAGE_CODE]
        i18n_titles = {
            language: factory.Faker("catch_phrase", locale=language).generate({})
            for language in languages
        }

    elif isinstance(title, dict):
        # Check that the languages passed are coherent with the languages requested if any
        if languages:
            assert set(languages).issubset(title.keys())
        else:
            languages = title.keys()
        i18n_titles = title

    elif isinstance(title, str):
        # Add a marker at the end of the string to differentiate each language
        languages = languages or [settings.LANGUAGE_CODE]
        i18n_titles = {language: title for language in languages}

    else:
        raise ValueError(
            "Title should be a string or a dictionary of language/string pairs"
        )

    # Assert that the languages passed are declared in settings
    assert set(languages).issubset({l[0] for l in settings.LANGUAGES})
    # Make a copy of languages to avoid muting it in what follows
    languages = list(languages)
    # Create the page with a first language from what is given to us
    first_language = languages.pop(0)

    slug = slugify(i18n_titles[first_language])
    page = create_page(
        language=first_language,
        menu_title=i18n_titles[first_language],
        title=i18n_titles[first_language],
        slug=slug,
        template=template,
        **kwargs,
    )

    if is_homepage is True:
        page.set_as_homepage()

    # Add a title for each additional language
    for language in languages:
        create_title(
            language=language,
            menu_title=i18n_titles[language],
            title=i18n_titles[language],
            slug=slugify(i18n_titles[language]),
            page=page,
        )
        # Publish page in each additional language
        if kwargs.get("published") is True:
            page.publish(language)

    return page


# pylint: disable=too-many-arguments
def create_text_plugin(
    page,
    slot,
    languages=None,
    is_html=True,
    max_nb_chars=None,
    nb_paragraphs=None,
    plugin_type="CKEditorPlugin",
):
    """
    A common function to create and add a text plugin of any type instance to
    a placeholder filled with some random text using Faker.

    Arguments:
        page (cms.models.pagemodel.Page): Instance of a Page used to search for
            given slot (aka a placeholder name).
        slot (string): A placeholder name available from page template.

    Keyword Arguments:
        languages (iterable): An iterable yielding language codes for which a text plugin should
            be created. If ``None`` (default) it uses the default language from settings.
        is_html (boolean): If True, every paragraph will be surrounded with an
            HTML paragraph markup. Default is True.
        max_nb_chars (integer): Number of characters limit to create each
            paragraph. Default is None so a random number between 200 and 400
            will be used at each paragraph.
        nb_paragraphs (integer): Number of paragraphs to create in content.
            Default is None so a random number between 2 and 4 will be used.
        plugin_type (string or object): Type of plugin. Default use CKEditorPlugin
            but you can use any other similar plugin that has a body attribute.

    Returns:
        object: Created plugin instance.
    """
    languages = languages or [settings.LANGUAGE_CODE]
    container = "<p>{:s}</p>" if is_html else "{:s}"
    nb_paragraphs = nb_paragraphs or random.randint(2, 4)

    placeholder = page.placeholders.get(slot=slot)

    for language in languages:
        paragraphs = []
        for _ in range(nb_paragraphs):
            max_nb_chars = max_nb_chars or random.randint(200, 400)
            paragraphs.append(
                factory.Faker(
                    "text", max_nb_chars=max_nb_chars, locale=language
                ).generate({})
            )
        body = [container.format(p) for p in paragraphs]

        add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type=plugin_type,
            body="".join(body),
        )


def recursive_page_creation(site, pages_info, parent=None):
    """
    Recursively create page following tree structure with parent/children.

    Arguments:
        site (django.contrib.sites.models.Site): Site object which page will
            be linked to.
        pages (dict): Page items to create recursively such as 'children' key
            value can be a dict to create child pages. The current page is
            given to children for parent relation.

    Keyword Arguments:
        parent (cms.models.pagemodel.Page): Page used as a parent to create
            page item from `pages` argument.

    Returns:
        dict: mapping of the page names passed in argument and the created page instances.
    """
    pages = {}

    for name, info in pages_info.items():
        page = create_i18n_page(
            info["title"],
            is_homepage=(name == "home"),
            in_navigation=info.get("in_navigation", True),
            published=True,
            site=site,
            parent=parent,
            **info["kwargs"],
        )

        pages[name] = page

        # Create children
        if info.get("children", None):
            children_pages = recursive_page_creation(
                site, info["children"], parent=page
            )
            for child_name in children_pages:
                if child_name in pages:
                    raise ImproperlyConfigured(
                        "Page names should be unique: {:s}".format(child_name)
                    )
            pages.update(children_pages)

    return pages



"""
Toolbar extension for the courses application
"""
from cms.toolbar_pool import toolbar_pool

from ..courses.cms_toolbars import BaseExtensionToolbar
from .models import Person


@toolbar_pool.register
class PersonExtensionToolbar(BaseExtensionToolbar):
    """
    This extension class customizes the toolbar for the person page extension
    """

    model = Person



"""Test utils for the courses application"""

from django.test.client import RequestFactory

from cms.middleware.toolbar import ToolbarMiddleware
from cms.toolbar.items import Menu, ModalItem


class CheckToolbarMixin:
    """
    Mixin that factorize `check_toolbar_item` method used by several applications
    CMS toolbar unit tests
    """

    def check_active(self, toolbar, name, item_type=ModalItem):
        """The item should be present in the toolbar and active."""
        page_menu_result = toolbar.find_items(Menu, name="Page")[0]
        results = page_menu_result.item.find_items(item_type, name=name)
        self.assertEqual(len(results), 1)
        item = results[0].item
        self.assertFalse(item.disabled)
        return item

    def check_disabled(self, toolbar, name, item_type=ModalItem):
        """The item should be present in the toolbar and disabled."""
        page_menu_result = toolbar.find_items(Menu, name="Page")[0]
        results = page_menu_result.item.find_items(item_type, name=name)
        self.assertEqual(len(results), 1)
        item = results[0].item
        self.assertTrue(item.disabled)
        return item

    def check_missing(self, toolbar, name, item_type=ModalItem):
        """The item should not be in the toolbar."""
        page_menu_result = toolbar.find_items(Menu, name="Page")[0]
        results = page_menu_result.item.find_items(item_type, name=name)
        self.assertEqual(results, [])

    # pylint: disable=unused-argument
    def check_absent(self, toolbar, *args, **kwargs):
        """The whole toolbar should be hidden."""
        self.assertEqual(toolbar.get_left_items(), [])

    @staticmethod
    def get_toolbar_for_page(page, user, edit, preview):
        """
        This method is a helper to build a request to test the toolbar in different states
        for different users
        """
        url = page.get_absolute_url()
        factory = RequestFactory()

        if edit:
            url = "{:s}?edit".format(url)
        else:
            url = "{:s}?edit_off".format(url)

        if preview:
            url = "{:s}&preview".format(url)

        request = factory.get(url)
        request.user = user
        request.current_page = page
        request.session = {}

        middleware = ToolbarMiddleware()
        middleware.process_request(request)

        # pylint: disable=no-member
        request.toolbar.get_left_items()
        return request.toolbar



"""
ElasticSearch category document management utilities
"""
from collections import defaultdict

from django.conf import settings
from django.db.models import Prefetch

from cms.models import Title
from djangocms_picture.models import Picture

from richie.plugins.simple_text_ckeditor.models import SimpleText

from ...courses.models import Category
from .. import defaults
from ..exceptions import QueryFormatException
from ..forms import CategoryListForm
from ..partial_mappings import MULTILINGUAL_TEXT
from ..utils.i18n import get_best_field_language
from ..utils.indexers import slice_string_for_completion


class CategoriesIndexer:
    """
    Makes available the parameters the indexer requires as well as functions to shape
    objects getting into and out of ElasticSearch
    """

    document_type = "category"
    index_name = "richie_categories"
    mapping = {
        "dynamic_templates": MULTILINGUAL_TEXT,
        "properties": {
            # Searchable
            **{
                "complete.{:s}".format(lang): {"type": "completion"}
                for lang, _ in settings.LANGUAGES
            },
            "description": {"type": "object"},
            "is_meta": {"type": "boolean"},
            "nb_children": {"type": "integer"},
            "path": {"type": "keyword"},
            "title": {"type": "object"},
            # Not searchable
            "absolute_url": {"type": "object", "enabled": False},
            "logo": {"type": "object", "enabled": False},
        },
    }
    scripts = {}
    display_fields = [
        "absolute_url",
        "is_meta",
        "logo",
        "nb_children",
        "path",
        "title.*",
    ]

    @classmethod
    def get_data_for_es(cls, index, action):
        """
        Load all the categories from the Category model and format them for the
        ElasticSearch index.
        """
        for category in (
            Category.objects.filter(
                extended_object__publisher_is_draft=False,
                extended_object__title_set__published=True,
            )
            .prefetch_related(
                Prefetch(
                    "extended_object__title_set",
                    to_attr="published_titles",
                    queryset=Title.objects.filter(published=True),
                )
            )
            .distinct()
        ):
            # Prepare published titles
            titles = {
                t.language: t.title for t in category.extended_object.published_titles
            }

            # Prepare logo images
            logo_images = {}
            for logo_image in Picture.objects.filter(
                cmsplugin_ptr__placeholder__page=category.extended_object,
                cmsplugin_ptr__placeholder__slot="logo",
            ):
                # Force the image format before computing it
                logo_image.use_no_cropping = False
                logo_image.width = defaults.CATEGORIES_LOGO_IMAGE_WIDTH
                logo_image.height = defaults.CATEGORIES_LOGO_IMAGE_HEIGHT
                logo_images[logo_image.cmsplugin_ptr.language] = logo_image.img_src

            # Prepare description texts
            description = defaultdict(list)
            for simple_text in SimpleText.objects.filter(
                cmsplugin_ptr__placeholder__page=category.extended_object,
                cmsplugin_ptr__placeholder__slot="description",
            ):
                description[simple_text.cmsplugin_ptr.language].append(simple_text.body)

            # Shorcut to the category's page node
            node = category.extended_object.node

            yield {
                "_id": str(category.extended_object_id),
                "_index": index,
                "_op_type": action,
                "_type": cls.document_type,
                "absolute_url": {
                    language: category.extended_object.get_absolute_url(language)
                    for language in titles.keys()
                },
                "complete": {
                    language: slice_string_for_completion(title)
                    for language, title in titles.items()
                },
                "description": {l: " ".join(st) for l, st in description.items()},
                "is_meta": bool(
                    node.parent is None
                    or node.parent.cms_pages.filter(category__isnull=True).exists()
                ),
                "logo": logo_images,
                "nb_children": node.numchild,
                "path": node.path,
                "title": titles,
            }

    @staticmethod
    def format_es_object_for_api(es_category, best_language):
        """
        Format an category stored in ES into a consistent and easy-to-consume record for
        API consumers
        """
        source = es_category["_source"]
        return {
            "id": es_category["_id"],
            "is_meta": source["is_meta"],
            "logo": get_best_field_language(source["logo"], best_language),
            "nb_children": source["nb_children"],
            "path": source["path"],
            "title": get_best_field_language(source["title"], best_language),
        }

    @staticmethod
    def build_es_query(request):
        """
        Build an ElasticSearch query and its related aggregations, to be consumed by the ES client
        in the Categories ViewSet
        """
        # Instantiate a form with our query_params to check & sanitize them
        params_form = CategoryListForm(request.query_params)

        # Raise an exception with error information if the query params are not valid
        if not params_form.is_valid():
            raise QueryFormatException(params_form.errors)

        # Build a query that matches on the name field if it was handed by the client
        # Note: test_elasticsearch_feature.py needs to be updated whenever the search call
        # is updated and makes use new features.
        if params_form.cleaned_data.get("query"):
            query = {
                "query": {
                    "match": {
                        "title.fr": {
                            "query": params_form.cleaned_data.get("query"),
                            "analyzer": "french",
                        }
                    }
                }
            }
        # Build a match_all query by default
        else:
            query = {"query": {"match_all": {}}}

        return (
            params_form.cleaned_data.get("limit"),
            params_form.cleaned_data.get("offset") or 0,
            query,
        )



"""
Core factories
"""
import io

from django.conf import settings
from django.contrib.auth.hashers import make_password
from django.core.files import File

import factory
from filer.models.imagemodels import Image

from ..core.helpers import create_i18n_page


class UserFactory(factory.django.DjangoModelFactory):
    """
    Create a fake user with Faker.
    """

    class Meta:
        model = settings.AUTH_USER_MODEL

    username = factory.Faker("user_name")
    email = factory.Faker("email")
    password = make_password("password")


class FilerImageFactory(factory.django.DjangoModelFactory):
    """
    Create a Filer Image for filer.fields.image.FilerImageField.
    """

    class Meta:
        model = Image

    owner = factory.SubFactory(UserFactory)
    original_filename = factory.Faker("file_name", category="image")

    # pylint: disable=no-self-use
    @factory.lazy_attribute
    def file(self):
        """
        Fill file field with generated image on the fly by PIL.

        Generated image is just a dummy blank image of 100x100 with plain blue
        color.

        Returns:
            django.core.files.File: File object.
        """
        # ImageField (both django's and factory_boy's) require PIL.
        # Try to import it along one of its known installation paths.
        try:
            from PIL import Image as PILimage
        except ImportError:
            import Image as PILimage

        thumb = PILimage.new("RGB", (100, 100), "blue")
        thumb_io = io.BytesIO()
        thumb.save(thumb_io, format="JPEG")

        return File(thumb_io, name=self.original_filename)


class PageExtensionDjangoModelFactory(factory.django.DjangoModelFactory):
    """
    Factories for page extensions have in common that:
    - they must create a related page with a title,
    - the related page may have to be placed below a "parent" page,
    - we may want to test the related page in several languages.

    All this is mutualized by inheriting from the present class.
    """

    @factory.lazy_attribute
    def extended_object(self):
        """
        Automatically create a related page with the title (or random title if None) in all the
        requested languages
        """
        return create_i18n_page(
            in_navigation=getattr(self, "page_in_navigation", False),
            languages=getattr(self, "page_languages", None),
            parent=getattr(self, "page_parent", None),
            template=getattr(self, "page_template", None),
            title=getattr(self, "page_title", None),
        )

    @classmethod
    def _after_postgeneration(cls, instance, create, results=None):
        """
        This hook method is called last when generating an instance from a factory. The super
        method saves the instance one last time after all the "post_generation" hooks have played.
        this is the moment to finally publish the pages. If we published the pages before this
        final "save", they would be set back to a pending state and would not be in a clean
        published state.
        """
        super()._after_postgeneration(instance, create, results=results)
        if results.get("should_publish", False):
            for language in instance.extended_object.get_languages():
                instance.extended_object.publish(language)
        instance.refresh_from_db()

    @factory.post_generation
    # pylint: disable=no-self-use,unused-argument
    def should_publish(self, create, extracted, **kwargs):
        """
        Mark the pages for publishing. The actual publishing is done by the
        "_after_post_generation" hook method above.
        """
        if create and extracted:
            return True
        return False



# Generated by Django 2.1.7 on 2019-02-15 17:04

import django.db.models.deletion
from django.db import migrations, models

import parler.models

import richie.apps.core.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [("cms", "0022_auto_20180620_1551")]

    operations = [
        migrations.CreateModel(
            name="Person",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "first_name",
                    models.CharField(max_length=200, verbose_name="First name"),
                ),
                (
                    "last_name",
                    models.CharField(max_length=200, verbose_name="Last name"),
                ),
                (
                    "extended_object",
                    models.OneToOneField(
                        editable=False,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="cms.Page",
                    ),
                ),
            ],
            options={"verbose_name": "person"},
        ),
        migrations.CreateModel(
            name="PersonPluginModel",
            fields=[
                (
                    "cmsplugin_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        related_name="persons_personpluginmodel",
                        serialize=False,
                        to="cms.CMSPlugin",
                    ),
                ),
                (
                    "page",
                    models.ForeignKey(
                        limit_choices_to={
                            "person__isnull": False,
                            "publisher_is_draft": True,
                        },
                        on_delete=django.db.models.deletion.CASCADE,
                        to="cms.Page",
                    ),
                ),
            ],
            options={"verbose_name": "person plugin model"},
            bases=(richie.apps.core.models.PagePluginMixin, "cms.cmsplugin"),
        ),
        migrations.CreateModel(
            name="PersonTitle",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                )
            ],
            options={"verbose_name": "person title"},
            bases=(parler.models.TranslatableModelMixin, models.Model),
        ),
        migrations.CreateModel(
            name="PersonTitleTranslation",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "language_code",
                    models.CharField(
                        db_index=True, max_length=15, verbose_name="Language"
                    ),
                ),
                ("title", models.CharField(max_length=200, verbose_name="Title")),
                (
                    "abbreviation",
                    models.CharField(max_length=10, verbose_name="Title abbreviation"),
                ),
                (
                    "master",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="translations",
                        to="persons.PersonTitle",
                    ),
                ),
            ],
            options={"verbose_name": "person title translation"},
        ),
        migrations.AddField(
            model_name="person",
            name="person_title",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="persons",
                to="persons.PersonTitle",
            ),
        ),
        migrations.AddField(
            model_name="person",
            name="public_extension",
            field=models.OneToOneField(
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="draft_extension",
                to="persons.Person",
            ),
        ),
        migrations.AlterUniqueTogether(
            name="persontitletranslation", unique_together={("language_code", "master")}
        ),
    ]



"""
Import custom settings and set up defaults for values the Search app needs
"""
from django.conf import settings
from django.forms import MultipleChoiceField
from django.utils.translation import gettext as _

import arrow

COURSES_COVER_IMAGE_WIDTH = getattr(settings, "COURSES_COVER_IMAGE_WIDTH", 216)
COURSES_COVER_IMAGE_HEIGHT = getattr(settings, "COURSES_COVER_IMAGE_HEIGHT", 216)

ORGANIZATIONS_LOGO_IMAGE_WIDTH = getattr(
    settings, "ORGANIZATIONS_LOGO_IMAGE_WIDTH", 216
)
ORGANIZATIONS_LOGO_IMAGE_HEIGHT = getattr(
    settings, "ORGANIZATIONS_LOGO_IMAGE_HEIGHT", 216
)

CATEGORIES_LOGO_IMAGE_WIDTH = getattr(settings, "CATEGORIES_LOGO_IMAGE_WIDTH", 216)
CATEGORIES_LOGO_IMAGE_HEIGHT = getattr(settings, "CATEGORIES_LOGO_IMAGE_HEIGHT", 216)

FILTERS_DEFAULT = [
    (
        "richie.apps.search.utils.filter_definitions.FilterDefinitionCustom",
        {
            "name": "new",
            "human_name": _("New courses"),
            "choices": [("new", _("First session"), [{"term": {"is_new": True}}])],
        },
    ),
    (
        "richie.apps.search.utils.filter_definitions.FilterDefinitionCustom",
        {
            "name": "availability",
            "human_name": _("Availability"),
            "choices": [
                (
                    "coming_soon",
                    _("Coming soon"),
                    [
                        {
                            "range": {
                                "start": {
                                    "gte": arrow.utcnow().datetime,
                                    "lte": arrow.utcnow().shift(weeks=+12).datetime,
                                }
                            }
                        }
                    ],
                ),
                (
                    "current",
                    _("Current session"),
                    [
                        {"range": {"start": {"lte": arrow.utcnow().datetime}}},
                        {"range": {"end": {"gte": arrow.utcnow().datetime}}},
                        {
                            "range": {
                                "enrollment_start": {"lte": arrow.utcnow().datetime}
                            }
                        },
                        {"range": {"enrollment_end": {"gte": arrow.utcnow().datetime}}},
                    ],
                ),
                (
                    "open",
                    _("Open for enrollment"),
                    [
                        {"range": {"start": {"lte": arrow.utcnow().datetime}}},
                        {"range": {"end": {"gte": arrow.utcnow().datetime}}},
                    ],
                ),
            ],
        },
    ),
    (
        "richie.apps.search.utils.filter_definitions.FilterDefinitionTerms",
        {"name": "categories", "human_name": _("Categories")},
    ),
    (
        "richie.apps.search.utils.filter_definitions.FilterDefinitionTerms",
        {"name": "organizations", "human_name": _("Organizations")},
    ),
    (
        "richie.apps.search.utils.filter_definitions.FilterDefinitionLanguages",
        {"name": "languages", "human_name": _("Languages")},
    ),
]
FILTERS = getattr(settings, "RICHIE_SEARCH_FILTERS", FILTERS_DEFAULT)


# Define our aggregations names, for our ES query, which will match with the field
# names on the objects & the facets we return on the API response
FILTERS_DYNAMIC = getattr(
    settings,
    "RICHIE_SEARCH_FILTERS_DYNAMIC",
    ["categories", "languages", "organizations"],
)

FILTERS_HARDCODED_DEFAULT = {
    filterParams["name"]: {
        "field": MultipleChoiceField,
        "choices": {
            choice_name: choice_fragment
            for choice_name, _, choice_fragment in filterParams["choices"]
        },
    }
    for className, filterParams in FILTERS_DEFAULT
    if "choices" in filterParams
}

FILTERS_HARDCODED = getattr(
    settings, "RICHIE_SEARCH_FILTERS_HARDCODED", FILTERS_HARDCODED_DEFAULT
)



"""
Persons application admin
"""

from django.contrib import admin

from cms.extensions import PageExtensionAdmin
from parler.admin import TranslatableAdmin

from .models import Person, PersonTitle


class PersonAdmin(PageExtensionAdmin):
    """Admin class for the Person model"""

    list_display = ["title", "person_title", "first_name", "last_name"]

    # pylint: disable=no-self-use
    def title(self, obj):
        """
        Display the person page title as a read-only field from the related page
        """
        return obj.extended_object.get_title()


class PersonTitleAdmin(TranslatableAdmin):
    """Admin class for the PersonTitle model"""

    list_display = ["title", "abbreviation"]


admin.site.register(Person, PersonAdmin)
admin.site.register(PersonTitle, PersonTitleAdmin)



"""
Define our default FilterDefinition classes.
They encapsulate common behavior around managing filters in ElasticSearch queries and
API requests (to validate input) and responses (to produce easy-to-consume filters).
"""
from functools import reduce

from django.conf import settings
from django.utils.module_loading import import_string

from .i18n import get_best_field_language


class FilterDefinitionBase:
    """
    Base filter definition shape: just an init method to set common attributes.
    """

    def __init__(self, name, human_name, is_drilldown=False):
        """
        Assign the common attributes and set some defaults for optional parameters.
        """
        self.name = name
        self.human_name = human_name
        self.is_drilldown = is_drilldown


class FilterDefinitionCustom(FilterDefinitionBase):
    """
    Filter definition for a custom filter with hardcoded choices provided along with their
    ES filtering fragment.
    """

    def __init__(self, name, human_name, choices, **kwargs):
        """
        Extend the base filter definition with tools to easily generate query/aggregation
        fragments and faceted filter definition.
        """
        super().__init__(name, human_name, **kwargs)
        # The key/fragment map is static and helpful to get query & aggregations fragments
        self._key_fragment_map = {choice[0]: choice[2] for choice in choices}
        # Extend the base filter definition with a list of values that includes names & human
        # names and is only missing the counts.
        # We do this because:
        #   - the number of choices is bounded as it is a static argument (as opposed to eg.)
        #     a row in the database;
        #   - it makes it easier to access the values by key with every search request if we
        #     use a dict (while keeping a simpler interface for richie lib users).
        self.values = {
            value_name: {"human_name": value_human_name, "key": value_name}
            for value_name, value_human_name, _ in choices
        }

    def get_query_fragment(self, value_list):
        """
        Build the query fragment to use in the ElasticSearch filter & aggregations.
        """
        # For custom filters, we pick the hardcoded fragment from the corresponding choices
        return [
            {"key": self.name, "fragment": self._key_fragment_map[value]}
            for value in value_list
        ]

    def get_aggs_fragment(self, queries):
        """
        Build the aggregations fragment to use to extract aggregations from ElasticSearch.
        """
        return {
            # Create a custom aggregation for each possible choice for this filter
            # eg `availability@coming_soon` & `availability@current` & `availability@open`
            "{:s}@{:s}".format(self.name, choice_key): {
                "filter": {
                    "bool": {
                        # Use all the query fragments from the queries *but* the one(s) that
                        # filter on the current filter: we manually add back the only one that
                        # is relevant to the current choice.
                        "must": choice_fragment
                        + [
                            # queries
                            # |> filter(kv_pair["key"] != self.name)
                            # |> map(pluck("fragment"))
                            # |> flatten()
                            clause
                            for kf_pair in queries
                            for clause in kf_pair["fragment"]
                            if kf_pair["key"] is not self.name
                        ]
                    }
                }
            }
            for choice_key, choice_fragment in self._key_fragment_map.items()
        }

    def get_faceted_definition(self, facets, _):
        """
        Simply add the counts to the values from the initial definition to make them complete
        as the frontend expects them.
        """
        return {
            # We always need to pass the base definition to the frontend
            "human_name": self.human_name,
            "is_drilldown": self.is_drilldown,
            "name": self.name,
            "values": [
                {
                    # Use the current value's name and human_name from the initial definition
                    # eg. for `availability@coming_soon` we pick self.values["comin_soon"]
                    **self.values[facet_name[facet_name.find("@") + 1 :]],  # noqa: E203
                    # Add the facet count
                    "count": facets[facet_name]["doc_count"],
                }
                # Operate from the facets as opposed to the filter definition to keep a more
                # flexible base that can support various ways to pick a subset of values for
                # which to show facets
                for facet_name in filter(
                    lambda fct_name: fct_name.startswith("{:s}@".format(self.name)),
                    facets,
                )
            ],
        }


class FilterDefinitionTerms(FilterDefinitionBase):
    """
    Filter definition for a terms-based filter. The choices are generated dynamically from
    the incoming facets to avoid having to hold in memory or iterate over an unbounded
    number of choices.
    """

    def get_query_fragment(self, value_list):
        """
        Build the query fragment to use in the ElasticSearch filter & aggregations.
        """
        # For terms filters, as the name implies, it's a simple terms fragment
        return [{"key": self.name, "fragment": [{"terms": {self.name: value_list}}]}]

    def get_aggs_fragment(self, queries):
        """
        Build the aggregations fragment to use to extract aggregations from ElasticSearch.
        """
        return {
            self.name: {
                # Rely on the built-in "terms" aggregation to get everything we need
                "aggregations": {self.name: {"terms": {"field": self.name}}},
                "filter": {
                    "bool": {
                        # Use all the query fragments from the queries *but* the one(s) that
                        # filter on the current filter, as it is handled by ElasticSearch for us
                        "must": [
                            # queries
                            # |> filter(kv_pair["key"] != self.name)
                            # |> map(pluck("fragment"))
                            # |> flatten()
                            clause
                            for kf_pair in queries
                            for clause in kf_pair["fragment"]
                            if kf_pair["key"] is not self.name
                        ]
                    }
                },
            }
        }

    def get_i18n_names(self, keys, language):
        """
        Helper method to get the corresponding internationalized human name for each key in
        a list of indexed objects' ids.
        This covers the base case for terms eg. other models in their own ElasticSearch index
        like organizations or categories
        """
        indexer = import_string(getattr(settings.ES_INDICES, self.name))

        # Get just the documents we need from ElasticSearch
        search_query_response = settings.ES_CLIENT.search(
            # We only need the titles to get the i18n names
            _source=["title.*"],
            index=indexer.index_name,
            doc_type=indexer.document_type,
            body={
                "query": {
                    "terms": {
                        "_uid": [
                            "{:s}#{:s}".format(indexer.document_type, key)
                            for key in keys
                        ]
                    }
                }
            },
            size=len(keys),
        )

        # Extract the best available language here to avoid handling these kinds of implementation
        # details in the ViewSet
        return {
            doc["_id"]: get_best_field_language(doc["_source"]["title"], language)
            for doc in search_query_response["hits"]["hits"]
        }

    def get_faceted_definition(self, facets, language):
        """
        Build the filter definition's values from base definition and the faceted keys.
        Those provide us with the keys and counts that we just have to consume. We resort to
        another method to get the internationalized names so specific filter definitions
        (ie subclasses) can have different ways to get those.
        """
        # Convert the keys & counts from ElasticSearch facets to a more readily consumable format
        #   {
        #       self.name: {
        #           self.name: {
        #               "buckets": [
        #                   {"key": "A", "count": x}
        #                   {"key": "B", "count": y}
        #                   {"key": "C", "count": z}
        #               ]
        #           }
        #       }
        #   }
        # 👆 becomes 👇
        #   {"A": x, "B": y, "C": z}
        key_count_map = reduce(
            lambda agg, key_count_dict: {
                **agg,
                key_count_dict["key"]: key_count_dict["doc_count"],
            },
            facets[self.name][self.name]["buckets"],
            {},
        )

        # Use the passed in callback to get internationalized names for all our keys
        key_i18n_name_map = self.get_i18n_names(
            [key for key in key_count_map], language
        )

        return {
            # We always need to pass the base definition to the frontend
            "human_name": self.human_name,
            "is_drilldown": self.is_drilldown,
            "name": self.name,
            "values": [
                # Aggregate the information from right above to build the values
                {"count": count, "human_name": key_i18n_name_map[key], "key": key}
                for key, count in key_count_map.items()
            ],
        }


class FilterDefinitionLanguages(FilterDefinitionTerms):
    """
    Languages need their own FilterDefinition subclass as there's a different way to get their
    human names from other terms-based filters
    """

    def get_i18n_names(self, keys, language):
        """
        All possible language keys should be in the ALL_LANGUAGES_DICT. We can just return it
        and let `get_faceted_definition` pick what it needs
        """
        return settings.ALL_LANGUAGES_DICT



"""
create_demo_site management command
"""
import logging
import random
from datetime import timedelta

from django.conf import settings
from django.contrib.sites.models import Site
from django.core.files import File
from django.core.management.base import BaseCommand, CommandError
from django.utils import timezone

from cms import models as cms_models
from cms.api import add_plugin
from filer.models.imagemodels import Image

from richie.apps.courses.factories import (
    VIDEO_SAMPLE_LINKS,
    CourseFactory,
    CourseRunFactory,
    LicenceFactory,
    OrganizationFactory,
)
from richie.apps.courses.helpers import create_categories
from richie.apps.courses.models import Category, Course, Licence, Organization
from richie.apps.persons.factories import (
    PersonFactory,
    PersonTitleFactory,
    PersonTitleTranslationFactory,
)
from richie.apps.persons.models import Person, PersonTitle, PersonTitleTranslation

from ...helpers import recursive_page_creation
from ...utils import file_getter

logger = logging.getLogger("richie.commands.core.create_demo_site")

DEMO_ANNEX_PAGE_ID = "annex"
NB_COURSES = 30
NB_COURSES_ORGANIZATION_RELATIONS = 3
NB_COURSES_PERSONS_PLUGINS = 3
NB_COURSES_SUBJECT_RELATIONS = 2
NB_ORGANIZATIONS = 5
NB_LICENCES = 5
NB_PERSONS = 10
NB_HOME_HIGHLIGHTED_COURSES = 8
NB_HOME_HIGHLIGHTED_ORGANIZATIONS = 4
NB_HOME_HIGHLIGHTED_SUBJECTS = 6
NB_HOME_HIGHLIGHTED_PERSONS = 3
PAGE_INFOS = {
    "home": {
        "title": {"en": "Home", "fr": "Accueil"},
        "in_navigation": False,
        "kwargs": {"template": "richie/homepage.html"},
    },
    "news": {
        "title": {"en": "News", "fr": "Actualités"},
        "in_navigation": True,
        "kwargs": {"template": "richie/fullwidth.html"},
    },
    "courses": {
        "title": {"en": "Courses", "fr": "Cours"},
        "in_navigation": True,
        "kwargs": {
            "reverse_id": Course.ROOT_REVERSE_ID,
            "template": "search/search.html",
        },
    },
    "categories": {
        "title": {"en": "Categories", "fr": "Catégories"},
        "in_navigation": False,
        "kwargs": {
            "reverse_id": Category.ROOT_REVERSE_ID,
            "template": "richie/child_pages_list.html",
        },
    },
    "organizations": {
        "title": {"en": "Organizations", "fr": "Etablissements"},
        "in_navigation": True,
        "kwargs": {
            "reverse_id": Organization.ROOT_REVERSE_ID,
            "template": "courses/cms/organization_list.html",
        },
    },
    "persons": {
        "title": {"en": "Persons", "fr": "Personnes"},
        "in_navigation": False,
        "kwargs": {
            "reverse_id": Person.ROOT_REVERSE_ID,
            "template": "richie/child_pages_list.html",
        },
    },
    "dashboard": {
        "title": {"en": "Dashboard", "fr": "Tableau de bord"},
        "in_navigation": False,
        "cms": False,
        "kwargs": {"template": "richie/fullwidth.html"},
    },
    "annex": {
        "title": {"en": "Annex", "fr": "Annexe"},
        "in_navigation": False,
        "kwargs": {
            "template": "richie/fullwidth.html",
            "reverse_id": DEMO_ANNEX_PAGE_ID,
        },
        "children": {
            "annex__about": {
                "title": {"en": "About", "fr": "A propos"},
                "in_navigation": True,
                "kwargs": {"template": "richie/fullwidth.html"},
            }
        },
    },
}
LEVELS_INFO = {
    "title": {"en": "Level", "fr": "Niveau"},
    "children": [
        {"title": {"en": "Beginner", "fr": "Débutant"}},
        {"title": {"en": "Advanced", "fr": "Avancé"}},
        {"title": {"en": "Expert", "fr": "Expert"}},
    ],
}
SUBJECTS_INFO = {
    "title": {"en": "Subject", "fr": "Subjet"},
    "children": [
        {
            "title": {"en": "Science", "fr": "Sciences"},
            "children": [
                {
                    "title": {
                        "en": "Agronomy and Agriculture",
                        "fr": "Agronomie et Agriculture",
                    }
                },
                {"title": {"en": "Chemistry", "fr": "Chimie"}},
                {
                    "title": {
                        "en": "Discovery of the Universe",
                        "fr": "Découverte de l'Univers",
                    }
                },
                {"title": {"en": "Environment", "fr": "Environnement"}},
                {
                    "title": {
                        "en": "Mathematics and Statistics",
                        "fr": "Mathématiques et Statistiques",
                    }
                },
                {
                    "title": {
                        "en": "Tools for Research",
                        "fr": "Outils pour la Recherche",
                    }
                },
                {"title": {"en": "Physics", "fr": "Physique"}},
                {"title": {"en": "Cognitive science", "fr": "Sciences cognitives"}},
                {
                    "title": {
                        "en": "Earth science and science of the Universe",
                        "fr": "Sciences de la Terre et de l'Univers",
                    }
                },
                {"title": {"en": "Life science", "fr": "Sciences de la vie"}},
                {
                    "title": {
                        "en": "Engineering science",
                        "fr": "Sciences pour l'ingénieur",
                    }
                },
            ],
        },
        {
            "title": {
                "en": "Human and social sciences",
                "fr": "Sciences humaines et social",
            },
            "children": [
                {"title": {"en": "Communication", "fr": "Communication"}},
                {
                    "title": {
                        "en": "Creation, Arts and Design",
                        "fr": "Création, Arts et Design",
                    }
                },
                {
                    "title": {
                        "en": "Culture and Civilization",
                        "fr": "Cultures et Civilisations",
                    }
                },
                {
                    "title": {
                        "en": "Social Issues and Social Policy",
                        "fr": "Enjeux de société",
                    }
                },
                {"title": {"en": "Geography", "fr": "Géographie"}},
                {"title": {"en": "History", "fr": "Histoire"}},
                {"title": {"en": "Innovation", "fr": "Innovation"}},
                {"title": {"en": "Literature", "fr": "Lettres"}},
                {"title": {"en": "Media", "fr": "Médias"}},
                {"title": {"en": "Philosophy", "fr": "Philosophie"}},
                {"title": {"en": "Political science", "fr": "Sciences politiques"}},
                {
                    "title": {
                        "en": "International relations",
                        "fr": "Relations internationales",
                    }
                },
                {"title": {"en": "Sports", "fr": "Sport"}},
            ],
        },
        {"title": {"en": "Law", "fr": "Droit et juridique"}},
        {"title": {"en": "Economy and Finance", "fr": "Economie et Finance"}},
        {"title": {"en": "Education and Training", "fr": "Education et formation"}},
        {"title": {"en": "Management", "fr": "Management"}},
        {"title": {"en": "Entrepreneurship", "fr": "Entreprenariat"}},
        {
            "title": {"en": "Computer science", "fr": "Informatique"},
            "children": [
                {
                    "title": {
                        "en": "Digital and Technology",
                        "fr": "Numérique et Technologie",
                    }
                },
                {
                    "title": {
                        "en": "Telecommunication and Networks",
                        "fr": "Télécommunications et Réseaux",
                    }
                },
                {"title": {"en": "Coding", "fr": "Programmation"}},
            ],
        },
        {"title": {"en": "Languages", "fr": "Langues"}},
        {"title": {"en": "Education and career guidance", "fr": "Orientation"}},
        {"title": {"en": "Health", "fr": "Santé"}},
    ],
}
HOMEPAGE_CONTENT = {
    "en": {
        "banner_title": "Welcome to Richie",
        "banner_content": "It works! This is the default homepage for the Richie CMS.",
        "banner_template": "richie/large_banner/hero-intro.html",
        "section_template": "richie/section/highlighted_items.html",
        "courses_title": "Popular courses",
        "organizations_title": "Universities",
        "persons_title": "Persons",
        "subjects_title": "Subjects",
    },
    "fr": {
        "banner_title": "Bienvenue sur Richie",
        "banner_content": "Ça marche ! Ceci est la page d'accueil par défaut du CMS Richie.",
        "banner_template": "richie/large_banner/hero-intro.html",
        "section_template": "richie/section/highlighted_items.html",
        "courses_title": "Cours à la une",
        "organizations_title": "Universités",
        "subjects_title": "Thématiques",
        "persons_title": "Personnes",
    },
}


def get_number_of_course_runs():
    """
    Returns a random integer between 1 and 5. We make it a convenience method so that it can
    be mocked in tests.
    """
    return random.randint(1, 5)


# pylint: disable=no-member
#
# Looks like pylint is not relevant at guessing object types when cascading
# methods over querysets: Instance of 'list' has no 'delete' member (no-member).
# We choose to ignore this false positive warning.


def clear_cms_data():
    """Clear all CMS data (CMS models + apps models)"""

    cms_models.Page.objects.all().delete()
    cms_models.Title.objects.all().delete()
    cms_models.CMSPlugin.objects.all().delete()
    cms_models.Placeholder.objects.all().delete()
    Course.objects.all().delete()  # Deletes the associated course runs as well by cascading
    Organization.objects.all().delete()
    Category.objects.all().delete()
    Person.objects.all().delete()
    PersonTitle.objects.all().delete()
    PersonTitleTranslation.objects.all().delete()
    Licence.objects.all().delete()


# pylint: disable=too-many-locals
def create_demo_site():
    """
    Create a simple site tree structure for developpers to work in realistic environment.

    We create multilingual pages, add organizations under the related page and add
    plugins to each page.
    """
    site = Site.objects.get(id=1)

    # Create pages as described in PAGES_INFOS
    pages_created = recursive_page_creation(site, PAGE_INFOS)

    # Create some licences
    licences = LicenceFactory.create_batch(NB_LICENCES)

    # Create organizations under the `Organizations` page
    organizations = OrganizationFactory.create_batch(
        NB_ORGANIZATIONS,
        page_in_navigation=True,
        page_languages=["en", "fr"],
        page_parent=pages_created["organizations"],
        fill_banner=True,
        fill_description=True,
        fill_logo=True,
        should_publish=True,
    )

    # Generate each category tree and return a list of the leaf categories
    levels = list(create_categories(LEVELS_INFO, pages_created["categories"]))
    subjects = list(create_categories(SUBJECTS_INFO, pages_created["categories"]))

    title = PersonTitleFactory(translation=None)
    PersonTitleTranslationFactory(
        master=title, language_code="en", title="Doctor", abbreviation="Dr."
    )
    PersonTitleTranslationFactory(
        master=title, language_code="fr", title="Docteur", abbreviation="Dr."
    )

    # Create persons under the `persons` page
    persons = PersonFactory.create_batch(
        NB_PERSONS,
        page_in_navigation=True,
        page_languages=["en", "fr"],
        page_parent=pages_created["persons"],
        person_title=random.choice([title, None]),
        fill_portrait=True,
        fill_resume=True,
        should_publish=True,
    )

    # Create courses under the `Course` page with categories and organizations
    # relations
    courses = []
    for _ in range(NB_COURSES):
        video_sample = random.choice(VIDEO_SAMPLE_LINKS)

        course = CourseFactory(
            page_in_navigation=True,
            page_languages=["en", "fr"],
            page_parent=pages_created["courses"],
            fill_licences=[
                ("course_license_content", random.choice(licences)),
                ("course_license_participation", random.choice(licences)),
            ],
            fill_team=random.sample(persons, NB_COURSES_PERSONS_PLUGINS),
            fill_teaser=video_sample,
            fill_cover=video_sample.image,
            fill_categories=[
                *random.sample(
                    subjects, random.randint(1, NB_COURSES_SUBJECT_RELATIONS)
                ),
                random.choice(levels),
            ],
            fill_organizations=random.sample(
                organizations, NB_COURSES_ORGANIZATION_RELATIONS
            ),
            fill_texts=[
                "course_syllabus",
                "course_format",
                "course_prerequisites",
                "course_plan",
                # "course_license_content",
                # "course_license_participation",
            ],
            should_publish=True,
        )
        courses.append(course)

        # Add a random number of course runs to the course
        nb_course_runs = get_number_of_course_runs()

        # 1) Make sure we have one course run open for enrollment
        now = timezone.now()
        CourseRunFactory(
            __sequence=nb_course_runs,
            page_in_navigation=False,
            page_parent=course.extended_object,
            start=now + timedelta(days=1),
            enrollment_start=now - timedelta(days=5),
            enrollment_end=now + timedelta(days=5),
            should_publish=True,
        )

        # 2) Add more random course runs
        for i in range(nb_course_runs - 1, 0, -1):
            CourseRunFactory(
                __sequence=i,
                page_in_navigation=False,
                page_languages=["en", "fr"],
                page_parent=course.extended_object,
                should_publish=True,
            )

    # Once everything has been created, use some content to create a homepage
    placeholder = pages_created["home"].placeholders.get(slot="maincontent")

    # - Get a banner image
    banner_file = file_getter("banner")()
    wrapped_banner = File(banner_file, banner_file.name)
    banner = Image.objects.create(file=wrapped_banner)

    # - Get a logo image
    logo_file = file_getter("logo")()
    wrapped_logo = File(logo_file, logo_file.name)
    logo = Image.objects.create(file=wrapped_logo)

    # - Create the home page in each language
    for language, content in HOMEPAGE_CONTENT.items():
        # Add a banner
        add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type="LargeBannerPlugin",
            title=content["banner_title"],
            background_image=banner,
            logo=logo,
            logo_alt_text="logo",
            content=content["banner_content"],
            template=content["banner_template"],
        )
        # Add highlighted courses
        courses_section = add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type="SectionPlugin",
            title=content["courses_title"],
            template=content["section_template"],
        )
        for course in random.sample(courses, NB_HOME_HIGHLIGHTED_COURSES):
            add_plugin(
                language=language,
                placeholder=placeholder,
                plugin_type="CoursePlugin",
                target=courses_section,
                page=course.extended_object,
            )

        # Add highlighted organizations
        organizations_section = add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type="SectionPlugin",
            title=content["organizations_title"],
            template=content["section_template"],
        )
        for organization in random.sample(
            organizations, NB_HOME_HIGHLIGHTED_ORGANIZATIONS
        ):
            add_plugin(
                language=language,
                placeholder=placeholder,
                plugin_type="OrganizationPlugin",
                target=organizations_section,
                page=organization.extended_object,
            )

        # Add highlighted subjects
        subjects_section = add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type="SectionPlugin",
            title=content["subjects_title"],
            template=content["section_template"],
        )
        for subject in random.sample(subjects, NB_HOME_HIGHLIGHTED_SUBJECTS):
            add_plugin(
                language=language,
                placeholder=placeholder,
                plugin_type="CategoryPlugin",
                target=subjects_section,
                page=subject.extended_object,
            )

        # Add highlighted persons
        persons_section = add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type="SectionPlugin",
            title=content["persons_title"],
            template=content["section_template"],
        )
        for person in random.sample(persons, NB_HOME_HIGHLIGHTED_PERSONS):
            add_plugin(
                language=language,
                placeholder=placeholder,
                plugin_type="PersonPlugin",
                target=persons_section,
                page=person.extended_object,
            )

        # Once content has been added we must publish again homepage in every
        # edited Languages
        pages_created["home"].publish("en")
        pages_created["home"].publish("fr")


class Command(BaseCommand):
    """Create default pages for FUN frontend"""

    help = __doc__

    def add_arguments(self, parser):

        parser.add_argument(
            "-f",
            "--force",
            action="store_true",
            default=False,
            help="Force command execution despite DEBUG is set to False",
        )

    def handle(self, *args, **options):

        if not settings.DEBUG and not options["force"]:
            raise CommandError(
                (
                    "This command is not meant to be used in production environment "
                    "except you know what you are doing, if so use --force parameter"
                )
            )

        clear_cms_data()
        create_demo_site()

        logger.info("done")



"""
Plain text CMS plugin
"""
from django.conf import settings
from django.utils.translation import ugettext_lazy as _

from cms.plugin_base import CMSPluginBase
from cms.plugin_pool import plugin_pool

from .models import PlainText


@plugin_pool.register_plugin
class PlainTextPlugin(CMSPluginBase):
    """
    A plugin to add plain text.
    """

    allow_children = False
    cache = True
    disable_child_plugins = True
    fieldsets = ((None, {"fields": ["body"]}),)
    model = PlainText
    module = settings.FUN_PLUGINS_GROUP
    name = _("Plain text")
    render_template = "richie/plain_text/plain_text.html"

    def render(self, context, instance, placeholder):
        """
        Build plugin context passed to its template to perform rendering
        """
        context = super(PlainTextPlugin, self).render(context, instance, placeholder)
        context["body"] = instance.body
        return context



"""
API endpoints to access courses through ElasticSearch
"""
from functools import reduce

from django.conf import settings
from django.utils.module_loading import import_string
from django.utils.translation import get_language_from_request

from elasticsearch.exceptions import NotFoundError
from rest_framework.response import Response
from rest_framework.viewsets import ViewSet

from ..defaults import FILTERS, FILTERS_DYNAMIC, FILTERS_HARDCODED
from ..exceptions import QueryFormatException
from ..utils.viewsets import AutocompleteMixin, ViewSetMetadata


class CoursesViewSet(AutocompleteMixin, ViewSet):
    """
    A simple viewset with GET endpoints to fetch courses
    See API Blueprint for details on consumer use
    """

    # Get the courses indexer from the settings
    _meta = ViewSetMetadata(indexer=import_string(settings.ES_INDICES.courses))

    # pylint: disable=no-self-use,unused-argument,too-many-locals
    def list(self, request, version):
        """
        Course search endpoint: build an ElasticSearch request from our query params so
        it searches its index and returns a list of matching courses
        """

        # Build the filters dict by instantiating all the active filters.
        # We do this once for every request to avoid instantiating them at every use site. Also,
        # it is not a good idea to build this during application bootstrap as category
        # filter definitions will need to be dynamically generated just in time.
        filters = {
            class_params["name"]: import_string(class_path)(**class_params)
            for class_path, class_params in FILTERS
        }

        try:
            limit, offset, query, aggs = self._meta.indexer.build_es_query(
                request, filters
            )
        except QueryFormatException as exc:
            # Return a 400 with error information if the query params are not as expected
            return Response(status=400, data={"errors": exc.args[0]})

        course_query_response = settings.ES_CLIENT.search(
            _source=getattr(self._meta.indexer, "display_fields", "*"),
            index=self._meta.indexer.index_name,
            doc_type=self._meta.indexer.document_type,
            body={
                "aggs": aggs,
                "query": query,
                "sort": self._meta.indexer.get_list_sorting_script(),
            },
            # Directly pass meta-params through as arguments to the ES client
            from_=offset,
            size=limit or settings.ES_DEFAULT_PAGE_SIZE,
        )

        response_object = {
            "meta": {
                "count": len(course_query_response["hits"]["hits"]),
                "offset": offset,
                "total_count": course_query_response["hits"]["total"],
            },
            "objects": [
                self._meta.indexer.format_es_object_for_api(
                    es_course,
                    # Get the best language we can return multilingual fields in
                    get_language_from_request(request),
                )
                for es_course in course_query_response["hits"]["hits"]
            ],
            # Transform facets (terms aggregations) into an easier-to-consume form: drop the meta
            # data and return objects with terms as keys and counts as values
            "facets": {
                # Resource (terms) facets are structured in the same way
                **{
                    field: {
                        term_value["key"]: term_value["doc_count"]
                        for term_value in value[field]["buckets"]
                    }
                    for field, value in course_query_response["aggregations"][
                        "all_courses"
                    ].items()
                    # Remove default fields inserted by elasticsearch
                    if field in FILTERS_DYNAMIC
                },
                # Custom (filter-based) facets are structured in another, common way
                **reduce(
                    # kv_pair is a tuple with (key, value) for our aggregations, eg. an item in
                    # [ (language@en, { doc_count: 42 }), (language@fr, { doc_count: 84 }) ]
                    # we want to output something way easier to consume and less redundant:
                    # { language: { en: 42, fr: 84 } }
                    lambda agg, kv_pair: {
                        # spread the existing aggregator to keep the already set aggs
                        **agg,
                        # Add or extend the object for one filter (eg. language, availability)
                        **dict(
                            [
                                (
                                    # Use the filter name as a key (eg. language, availability)
                                    kv_pair[0][: kv_pair[0].find("@")],
                                    {
                                        # Keep the key/value pairs already on the aggregator
                                        **agg.get(
                                            kv_pair[0][: kv_pair[0].find("@")], {}
                                        ),
                                        # Add our incoming key/value pair (eg. en: 84)
                                        **dict(
                                            [
                                                (
                                                    # Use the filter value as a key (eg. fr, open)
                                                    kv_pair[0][
                                                        kv_pair[0].find("@")
                                                        + 1 :  # noqa: E203
                                                    ],
                                                    # The actual interesting value is the count
                                                    kv_pair[1]["doc_count"],
                                                )
                                            ]
                                        ),
                                    },
                                )
                            ]
                        ),
                    },
                    # Select only the aggregations coming from hardcoded filters
                    # kv_pair is a tuple with (key, value) for our aggregations (see above)
                    filter(
                        lambda kv_pair: kv_pair[0][: kv_pair[0].find("@")]
                        in FILTERS_HARDCODED,
                        course_query_response["aggregations"]["all_courses"].items(),
                    ),
                    {},
                ),
            },
            "filters": {
                filter.name: filter.get_faceted_definition(
                    course_query_response["aggregations"]["all_courses"],
                    get_language_from_request(request),
                )
                for filter in filters.values()
            },
        }

        # Will be formatting a response_object for consumption
        return Response(response_object)

    # pylint: disable=no-self-use,invalid-name,unused-argument
    def retrieve(self, request, pk, version):
        """
        Return a single course by ID
        """
        # Wrap the ES get in a try/catch to we control the exception we emit — it would
        # raise and end up in a 500 error otherwise
        try:
            query_response = settings.ES_CLIENT.get(
                index=self._meta.indexer.index_name,
                doc_type=self._meta.indexer.document_type,
                id=pk,
            )
        except NotFoundError:
            return Response(status=404)

        # Format a clean course object as a response
        return Response(
            self._meta.indexer.format_es_object_for_api(
                query_response,
                # Get the best language we can return multilingual fields in
                get_language_from_request(request),
            )
        )



"""
ElasticSearch course document management utilities
"""
from collections import defaultdict
from datetime import MAXYEAR
from functools import reduce

from django.conf import settings
from django.db.models import Prefetch

import arrow
import dateutil.parser
from cms.models import Title
from djangocms_picture.models import Picture

from richie.plugins.simple_text_ckeditor.models import SimpleText

from ...courses.models import Course, CourseRun
from ..defaults import COURSES_COVER_IMAGE_HEIGHT, COURSES_COVER_IMAGE_WIDTH
from ..exceptions import QueryFormatException
from ..forms import CourseListForm
from ..partial_mappings import MULTILINGUAL_TEXT
from ..utils.i18n import get_best_field_language
from ..utils.indexers import slice_string_for_completion


class CoursesIndexer:
    """
    Makes available the parameters the indexer requires as well as functions to shape
    objects getting into and out of ElasticSearch
    """

    document_type = "course"
    index_name = "richie_courses"
    mapping = {
        "dynamic_templates": MULTILINGUAL_TEXT,
        "properties": {
            # Dates
            "start": {"type": "date"},
            "end": {"type": "date"},
            "enrollment_start": {"type": "date"},
            "enrollment_end": {"type": "date"},
            # Keywords
            "categories": {"type": "keyword"},
            "languages": {"type": "keyword"},
            "organizations": {"type": "keyword"},
            # Searchable
            **{
                "complete.{:s}".format(lang): {"type": "completion"}
                for lang, _ in settings.LANGUAGES
            },
            "description": {"type": "object"},
            "is_new": {"type": "boolean"},
            "title": {"type": "object"},
            # Not searchable
            "absolute_url": {"type": "object", "enabled": False},
            "cover_image": {"type": "object", "enabled": False},
        },
    }
    display_fields = [
        "start",
        "end",
        "enrollment_start",
        "enrollment_end",
        "absolute_url",
        "cover_image",
        "languages",
        "organizations",
        "categories",
        "title.*",
    ]

    # Define the scoring boost (in ElasticSearch) related value names receive when using
    # full-text search.
    # For example, when a user searches for "Science" in full-text, it should match any
    # course whose category contains "Science" or a related word, albeit with a lower
    # score than courses that include it in their title or description.
    # This lower score factor is the boost value we get or set here.
    fulltext_search_filter_matching_boost = 0.05

    scripts = {
        # The ordering process first splits the courses into four groups, with further ordering
        # inside each one of those groups.
        #
        # Here's a schematic representation that shows the ordering factor for each
        # course (to be used in ascending order later on) :
        #
        #          TOP OF THE LIST
        #          ———————————————
        # ----- NOW (current timestamp) -----
        #         Courses in bucket 1
        #
        #  < ~1 x datetime.MAXYEAR distance >
        #
        # ------ ~1 x datetime.MAXYEAR ------
        #         Courses in bucket 2
        #
        #  < ~1 x datetime.MAXYEAR distance >
        #
        #         Courses in bucket 3
        # ------ ~2 x datetime.MAXYEAR ------
        #
        #  < ~1 x datetime.MAXYEAR distance >
        #
        #         Courses in bucket 4
        # ------ ~3 x datetime.MAXYEAR ------
        #          ———————————————
        #          END OF THE LIST
        #
        # For reference MAXYEAR's timestamp is more than 2 orders of magnitude larger than
        # this year's timestamp (2018).
        # This means there can be no overlap between the various buckets, but we can still
        # sort courses inside each bucket as we see fit by simply adding timestamps (ascending
        # order) or substracting them (descending order).
        "sort_list": {
            "script": {
                "lang": "expression",
                "source": (
                    # 4- Courses that have ended.
                    # Ordered by descending end datetime. The course that has finished last
                    # is displayed first.
                    "doc['end'].value < ms_since_epoch ? "
                    "3 * max_date - doc['end'].value : "
                    # 3- Courses that have not ended but can no longer be enrolled in.
                    # Ordered by descending end of enrollment datetime. The course for which
                    # enrollment has ended last is displayed first.
                    "doc['enrollment_end'].value < ms_since_epoch ? "
                    "2 * max_date - doc['enrollment_end'].value : "
                    # 2- Courses that have not started yet.
                    # Ordered by starting datetime. The next course to start is displayed first.
                    "ms_since_epoch < doc['start'].value ? "
                    "max_date + doc['start'].value : "
                    # 1- Courses that are currently open and can be enrolled in.
                    # Ordered by ascending end of enrollment datetime. The next course to end
                    # enrollment is displayed first.
                    "doc['enrollment_end'].value"
                ),
            }
        }
    }

    @classmethod
    def get_data_for_es(cls, index, action):
        """
        Load all the course runs from the Course model and format them for the ElasticSearch index
        """
        for course in (
            Course.objects.filter(
                extended_object__publisher_is_draft=False,
                extended_object__title_set__published=True,
            )
            .prefetch_related(
                Prefetch(
                    "extended_object__title_set",
                    to_attr="published_titles",
                    queryset=Title.objects.filter(published=True),
                )
            )
            .distinct()
        ):
            # Prepare published titles
            titles = {
                t.language: t.title for t in course.extended_object.published_titles
            }

            # Prepare cover images
            cover_images = {}
            for cover_image in Picture.objects.filter(
                cmsplugin_ptr__placeholder__page=course.extended_object,
                cmsplugin_ptr__placeholder__slot="course_cover",
            ):
                # Force the image format before computing it
                cover_image.use_no_cropping = False
                cover_image.width = COURSES_COVER_IMAGE_WIDTH
                cover_image.height = COURSES_COVER_IMAGE_HEIGHT
                cover_images[cover_image.cmsplugin_ptr.language] = cover_image.img_src

            # Prepare syllabus texts
            syllabus_texts = defaultdict(list)
            for simple_text in SimpleText.objects.filter(
                cmsplugin_ptr__placeholder__page=course.extended_object,
                cmsplugin_ptr__placeholder__slot="course_syllabus",
            ):
                syllabus_texts[simple_text.cmsplugin_ptr.language].append(
                    simple_text.body
                )

            # Make sure we get title information for categories in the same request
            category_pages = (
                course.get_root_to_leaf_category_pages()
                .prefetch_related(
                    Prefetch(
                        "title_set",
                        to_attr="published_titles",
                        queryset=Title.objects.filter(published=True),
                    )
                )
                .only("pk")
            )

            # Make sure we get title information for organizations in the same request
            organizations = (
                course.get_organizations()
                .prefetch_related(
                    Prefetch(
                        "extended_object__title_set",
                        to_attr="published_titles",
                        queryset=Title.objects.filter(published=True),
                    )
                )
                .only("extended_object")
                .distinct()
            )

            course_runs = course.get_course_runs()
            for course_run in course_runs:
                yield {
                    "_id": str(course_run.extended_object_id),
                    "_index": index,
                    "_op_type": action,
                    "_type": cls.document_type,
                    "start": course_run.start,
                    "end": course_run.end,
                    "enrollment_start": course_run.enrollment_start,
                    "enrollment_end": course_run.enrollment_end,
                    "absolute_url": {
                        language: course_run.extended_object.get_absolute_url(language)
                        for language in titles.keys()
                    },
                    "categories": [str(page.pk) for page in category_pages],
                    # Index the names of categories to surface them in full text searches
                    "categories_names": reduce(
                        lambda acc, title: {
                            **acc,
                            title.language: acc[title.language] + [title.title]
                            if acc.get(title.language)
                            else [title.title],
                        },
                        [
                            title
                            for page in category_pages
                            for title in page.published_titles
                        ],
                        {},
                    ),
                    "complete": {
                        language: slice_string_for_completion(title)
                        for language, title in titles.items()
                    },
                    "cover_image": cover_images,
                    "description": {
                        l: " ".join(st) for l, st in syllabus_texts.items()
                    },
                    "is_new": len(course_runs) == 1,
                    "languages": course_run.languages,
                    "organizations": [
                        str(id)
                        for id in course.get_organizations().values_list(
                            "public_extension__extended_object", flat=True
                        )
                        if id is not None
                    ],
                    # Index the names of organizations to surface them in full text searches
                    "organizations_names": reduce(
                        lambda acc, title: {
                            **acc,
                            title.language: acc[title.language] + [title.title]
                            if acc.get(title.language)
                            else [title.title],
                        },
                        [
                            title
                            for organization in organizations
                            for title in organization.extended_object.published_titles
                        ],
                        {},
                    ),
                    "title": titles,
                }

    @staticmethod
    def format_es_object_for_api(es_course, best_language):
        """
        Format a course stored in ES into a consistent and easy-to-consume record for
        API consumers
        """
        source = es_course["_source"]
        return {
            "id": es_course["_id"],
            "start": source["start"],
            "end": source["end"],
            "enrollment_start": source["enrollment_start"],
            "enrollment_end": source["enrollment_end"],
            "absolute_url": get_best_field_language(
                source["absolute_url"], best_language
            ),
            "categories": source["categories"],
            "cover_image": get_best_field_language(
                source["cover_image"], best_language
            ),
            "languages": source["languages"],
            "organizations": source["organizations"],
            "state": CourseRun.compute_state(
                dateutil.parser.parse(source["start"]),
                dateutil.parser.parse(source["end"]),
                dateutil.parser.parse(source["enrollment_start"]),
                dateutil.parser.parse(source["enrollment_end"]),
            )._asdict(),
            "title": get_best_field_language(source["title"], best_language),
        }

    @classmethod
    # pylint: disable=R0912, R0914
    def build_es_query(cls, request, filters):
        """
        Build an ElasticSearch query and its related aggregations, to be consumed by the ES client
        in the Courses ViewSet
        """
        # QueryDict/MultiValueDict breaks lists: we need to normalize them
        # Unpacking does not trigger the broken accessor so we get the proper value
        params_form_values = {
            k: v[0] if len(v) == 1 else v for k, v in request.query_params.lists()
        }
        # Use QueryDict/MultiValueDict as a shortcut to make sure we get arrays for our filters,
        # which should be arrays even if their length is one
        for filter_name in filters:
            params_form_values[filter_name] = request.query_params.getlist(filter_name)
        # Instantiate the form to allow validation/cleaning
        params_form = CourseListForm(params_form_values)

        # Raise an exception with error information if the query params are not valid
        if not params_form.is_valid():
            raise QueryFormatException(params_form.errors)

        # Note: test_elasticsearch_feature.py needs to be updated whenever the search call
        # is updated and makes use new features.
        # queries is an array of individual queries that will be combined through "bool" before
        # we pass them to ES. See the docs en bool queries.
        # https://www.elastic.co/guide/en/elasticsearch/reference/current/query-dsl-bool-query.html
        queries = []
        for param, value in params_form.cleaned_data.items():
            # Skip falsy values as we're not using them in our query
            if not value:
                continue

            # The datetimerange fields are all translated to the ES query DSL the same way
            if param in ["end", "enrollment_end", "enrollment_start", "start"]:
                # Add the relevant range criteria to the queries
                start, end = value
                queries = [
                    *queries,
                    {
                        "key": param,
                        "fragment": [
                            {
                                "range": {
                                    param: {
                                        "gte": start.datetime if start else None,
                                        "lte": end.datetime if end else None,
                                    }
                                }
                            }
                        ],
                    },
                ]

            elif param in filters:
                # Add the query fragments to the query
                queries = queries + filters[param].get_query_fragment(value)

            # Search is a regular (multilingual) match query
            elif param == "query":
                queries = [
                    *queries,
                    {
                        "key": param,
                        "fragment": [
                            {
                                "bool": {
                                    "should": [
                                        {
                                            "multi_match": {
                                                "fields": ["description.*", "title.*"],
                                                "query": value,
                                                "type": "cross_fields",
                                            }
                                        },
                                        {
                                            "multi_match": {
                                                "boost": cls.fulltext_search_filter_matching_boost,
                                                "fields": [
                                                    "categories_names.*",
                                                    "organizations_names.*",
                                                ],
                                                "query": value,
                                                "type": "cross_fields",
                                            }
                                        },
                                    ]
                                }
                            }
                        ],
                    },
                ]

        # Default to a match_all query
        if not queries:
            query = {"match_all": {}}
        else:
            # Concatenate all the sub-queries lists together to form the queries list
            query = {
                "bool": {
                    "must":
                    # queries => map(pluck("fragment")) => flatten()
                    [clause for kf_pair in queries for clause in kf_pair["fragment"]]
                }
            }

        # Concatenate our hardcoded filters query fragments with organizations and categories terms
        # aggregations build on-the-fly
        aggs = {
            "all_courses": {
                "global": {},
                "aggregations": reduce(
                    # Merge all the partial aggregations dicts together
                    lambda acc, aggs_fragment: {**acc, **aggs_fragment},
                    # Generate a partial aggregations dict (an aggs_fragment) for each filter
                    [filter.get_aggs_fragment(queries) for filter in filters.values()],
                    {},
                ),
            }
        }

        return (
            params_form.cleaned_data.get("limit"),
            params_form.cleaned_data.get("offset") or 0,
            query,
            aggs,
        )

    @staticmethod
    def get_list_sorting_script():
        """
        Call the relevant sorting script for courses lists, regenerating the parameters on each
        call. This will allow the ms_since_epoch value to stay relevant even if the ES instance
        and/or the Django server are long running.

        Note: we use script storage to save time on the script compilation, which is an expensive
        operation. We'll only do it once at bootstrap time.
        """
        return {
            "_script": {
                "order": "asc",
                "script": {
                    "id": "sort_list",
                    "params": {
                        "max_date": arrow.get(MAXYEAR, 12, 31).timestamp * 1000,
                        "ms_since_epoch": arrow.utcnow().timestamp * 1000,
                    },
                },
                "type": "number",
            }
        }



"""
Helpers that can be useful throughout Richie's courses app
"""
from .factories import CategoryFactory


def create_categories(info, parent, should_publish=True):
    """
    Create the category tree from the SUBJECTS dictionary.


    Arguments:
        info (List): definition of the category tree to create in the following format:

            {
                "title": "Subject",
                "children": [
                    {
                        "title": "Computer science",
                        "children": [
                            {"title": "Coding"},
                            {"title": "Security"},
                        ],
                    },
                    {"title": "Languages"},
                ],
            }

        page (cms.models.pagemodel.Page): Instance of a Page below which the category
            tree is created.

    Returns:
        generator[courses.models.Category]: yield only the leaf categories of the created tree.

    """
    category = CategoryFactory(
        page_title=info["title"],
        page_in_navigation=info.get("in_navigation", True),
        page_parent=parent,
        should_publish=should_publish,
    )

    if info.get("children", None):
        for child_info in info["children"]:
            yield from create_categories(
                child_info, category.extended_object, should_publish=should_publish
            )
    else:
        # we only return leaf categories (no children)
        yield category
